import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

import eradiate
import eradiate.experiments as ertxp
import eradiate.scenes as ertsc
from eradiate.units import symbol
from eradiate.units import unit_registry as ureg


@pytest.mark.slow
@pytest.mark.parametrize("reflectance", [0.0, 0.5, 1.0])
@pytest.mark.parametrize("sza", [0.0, 30.0, 60.0])
def test_compare_canopy_atmosphere_vs_atmosphere(
    mode_ckd_double,
    reflectance,
    sza,
    artefact_dir,
    ert_seed_state,
    atmosphere_us_standard_ckd,
):
    r"""
    Compare CanopyAtmosphereExperiment with AtmosphereExperiment
    ============================================================

    This test compares the results of an AtmosphereExperiment with a
    corresponding CanopyAtmosphereExperiment run. To achieve that, the
    CanopyAtmosphereExperiment is set up without an explicit canopy and a basic
    atmosphere.

    Rationale
    ---------

    Both experiments are set up with the same surface and atmosphere:

    * Geometry: a square surface with unit size and a Lambertian BRDF with
      reflectance :math:`\rho \in [0.0, 0.5, 1.0]`.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta \in [0.0, 30.0, 60.0]°`.
    * Atmosphere: Molecular atmosphere using the ussa1976 profile
    * Sensor: Distant reflectance measure covering a plane (21 angular points,
      10000 sample per pixel)
      We take an odd number of angular points to ensure that the special value
      :math:`\theta = 0°` is included.

    Expected behaviour
    ------------------

    The results of the two runs must be exactly equal.

    Results
    -------

    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_0.0_0.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_0.0_0.5.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_0.0_1.0.png
       :width: 30%

    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_30.0_0.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_30.0_0.5.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_30.0_1.0.png
       :width: 30%

    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_60.0_0.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_60.0_0.5.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_atmosphere_60.0_1.0.png
       :width: 30%
    """

    surface = {"type": "lambertian", "reflectance": reflectance}
    illumination = {"type": "directional", "zenith": sza, "azimuth": 0.0}
    molecular_atmosphere = atmosphere_us_standard_ckd
    molecular_atmosphere.update({"has_absorption": False})
    atmosphere = {
        "type": "heterogeneous",
        "molecular_atmosphere": molecular_atmosphere,
    }
    measure = {
        "type": "distant",
        "id": "measure",
        "construct": "hplane",
        "zeniths": np.arange(-90, 91, 15),
        "azimuth": 0.0,
        "spp": 1e4,
        "srf": ertsc.spectra.MultiDeltaSpectrum(
            wavelengths=550 * ureg.nm,
        ),
    }

    onedim = ertxp.AtmosphereExperiment(
        surface=surface,
        illumination=illumination,
        atmosphere=atmosphere,
        measures=measure,
    )

    r4a = ertxp.CanopyAtmosphereExperiment(
        surface=surface,
        illumination=illumination,
        atmosphere=atmosphere,
        measures=measure,
        canopy=None,
    )

    ert_seed_state.reset()
    eradiate.run(onedim, seed_state=ert_seed_state)

    ert_seed_state.reset()
    eradiate.run(r4a, seed_state=ert_seed_state)

    fig = plt.figure(figsize=(6, 3))
    ax = plt.gca()

    onedim_vza = np.squeeze(onedim.results["measure"].vza.values)
    onedim_radiance = np.squeeze(onedim.results["measure"]["radiance"].values)
    ax.plot(onedim_vza, onedim_radiance, label="onedim", marker=".", ls="--")

    r4a_vza = np.squeeze(r4a.results["measure"].vza.values)
    r4a_radiance = np.squeeze(r4a.results["measure"]["radiance"].values)
    ax.plot(r4a_vza, r4a_radiance, label="r4a", marker=".", ls="--")

    radiance_units = symbol(r4a.results["measure"]["radiance"].attrs["units"])
    plt.xlabel("Signed viewing zenith angle [°]")
    plt.xticks([-90.0, -60.0, -30.0, 0.0, 30.0, 60.0, 90.0])
    plt.ylabel(f"Radiance [{radiance_units}]")
    plt.title(rf"SZA = {sza} — $\rho$ = {reflectance}")
    plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    filename = f"test_compare_canopy_atmosphere_vs_atmosphere_{sza}_{reflectance}.png"
    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    fname_plot = os.path.join(outdir, filename)
    plt.tight_layout()
    fig.savefig(fname_plot, dpi=200)
    plt.close()

    assert np.all(
        onedim.results["measure"]["radiance"].values
        == r4a.results["measure"]["radiance"].values
    )


@pytest.mark.slow
@pytest.mark.parametrize("sza", [0.0, 30.0, 60.0])
@pytest.mark.parametrize("lai", [1.0, 2.0, 3.0])
def test_compare_canopy_atmosphere_vs_canopy(
    mode_mono_double, sza, lai, artefact_dir, ert_seed_state
):
    r"""
    Compare CanopyAtmosphereExperiment with CanopyExperiment
    ========================================================

    This test compares the results of a CanopyExperiment with a corresponding
    CanopyAtmosphereExperiment run. To achieve that, the
    CanopyAtmosphereExperiment is set up without an atmosphere.

    Rationale
    ---------

    Both experiments are set up with the same surface and atmosphere:

    * Geometry: a square surface with unit size and a Lambertian BRDF with
      reflectance :math:`\rho = 0.25`.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta \in [0.0, 30.0, 60.0]°`.
    * Atmosphere: None
    * Canopy: Homogeneous discrete canopy with LAI :math:`[1.0, 2.0, 3.0]`
    * Sensor: Distant reflectance measure covering a plane (21 angular points,
      10000 sample per pixel)
      We take an odd number of angular points to ensure that the special value
      :math:`\theta = 0°` is included.

    Expected behaviour
    ------------------

    The results of the two runs must be exactly equal.

    Results
    -------

    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_0.0_1.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_0.0_2.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_0.0_3.0.png
       :width: 30%

    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_30.0_1.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_30.0_2.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_30.0_3.0.png
       :width: 30%

    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_60.0_1.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_60.0_2.0.png
       :width: 30%
    .. image:: generated/plots/test_compare_canopy_atmosphere_vs_canopy_60.0_3.0.png
       :width: 30%
    """

    surface = {"type": "lambertian", "reflectance": 0.25}
    illumination = {"type": "directional", "zenith": sza, "azimuth": 0.0}
    canopy = ertsc.biosphere.DiscreteCanopy.homogeneous(
        lai=lai,
        leaf_radius=0.1 * ureg.m,
        l_horizontal=10.0 * ureg.m,
        l_vertical=2.0 * ureg.m,
        padding=5,
    )
    measure = {
        "type": "distant",
        "id": "measure",
        "construct": "hplane",
        "zeniths": np.arange(-90, 91, 15),
        "azimuth": 0.0,
        "spp": 1e4,
        "srf": ertsc.spectra.MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm),
    }

    rami = ertxp.CanopyExperiment(
        surface=surface,
        illumination=illumination,
        canopy=canopy,
        measures=measure,
    )

    r4a = ertxp.CanopyAtmosphereExperiment(
        surface=surface,
        illumination=illumination,
        atmosphere=None,
        canopy=canopy,
        measures=measure,
    )

    ert_seed_state.reset()
    eradiate.run(rami, seed_state=ert_seed_state)

    ert_seed_state.reset()
    eradiate.run(r4a, seed_state=ert_seed_state)

    fig = plt.figure(figsize=(6, 3))
    ax = plt.gca()

    rami_vza = np.squeeze(rami.results["measure"].vza.values)
    rami_radiance = np.squeeze(rami.results["measure"]["radiance"].values)
    ax.plot(rami_vza, rami_radiance, label="rami", marker=".", ls="--")

    r4a_vza = np.squeeze(r4a.results["measure"].vza.values)
    r4a_radiance = np.squeeze(r4a.results["measure"]["radiance"].values)
    ax.plot(r4a_vza, r4a_radiance, label="r4a", marker=".", ls="--")

    radiance_units = symbol(r4a.results["measure"]["radiance"].attrs["units"])
    plt.xlabel("Signed viewing zenith angle [°]")
    plt.xticks([-90.0, -60.0, -30.0, 0.0, 30.0, 60.0, 90.0])
    plt.ylabel(f"Radiance [{radiance_units}]")
    plt.title(f"SZA = {sza} — LAI = {lai}")
    plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    filename = f"test_compare_canopy_atmosphere_vs_canopy_{sza}_{lai}.png"
    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    fname_plot = os.path.join(outdir, filename)
    plt.tight_layout()
    fig.savefig(fname_plot, dpi=200)
    plt.close()

    assert np.all(
        rami.results["measure"]["radiance"].values
        == r4a.results["measure"]["radiance"].values
    )



import mitsuba as mi
import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.measure import (
    AngleLayout,
    AzimuthRingLayout,
    DirectionLayout,
    GridLayout,
    HemispherePlaneLayout,
    Layout,
    MultiDistantMeasure,
    measure_factory,
)
from eradiate.test_tools.types import check_scene_element

# ------------------------------------------------------------------------------
#                               Layout framework
# ------------------------------------------------------------------------------


def test_angle_layout(mode_mono):
    """
    Unit tests for the AngleLayout class
    """
    # Constructing without argument fails
    with pytest.raises(TypeError):
        AngleLayout()

    # (2,) arrays are reshaped as (1, 2)
    assert AngleLayout([0, 0] * ureg.deg).angles.shape == (1, 2)

    # Zenith values outside [0, 180]° are not allowed
    with pytest.raises(ValueError):
        AngleLayout([-45, 0] * ureg.deg)
    with pytest.raises(ValueError):
        AngleLayout([210, 0] * ureg.deg)

    # Regular construction pattern succeeds
    layout = AngleLayout([[0, 0], [45, 0], [45, 90], [45, 180]] * ureg.deg)

    # Directions are correctly computed and point outwards
    np.testing.assert_allclose(
        layout.directions,
        [
            [0, 0, 1],
            [np.sqrt(2) / 2, 0, np.sqrt(2) / 2],
            [0, np.sqrt(2) / 2, np.sqrt(2) / 2],
            [-np.sqrt(2) / 2, 0, np.sqrt(2) / 2],
        ],
        atol=1e-8,
    )

    # Dictionary conversion protocol applies units
    layout = Layout.convert(
        {
            "type": "angles",
            "angles": [
                [0, 0],
                [np.pi / 4, 0],
                [np.pi / 4, np.pi / 2],
                [np.pi / 4, np.pi],
            ],
            "angles_units": "rad",
        }
    )
    np.testing.assert_allclose(
        layout.angles.m_as(ureg.deg),
        [[0, 0], [45, 0], [45, 90], [45, 180]],
        atol=1e-8,
    )


@pytest.mark.parametrize(
    "convention, expected",
    [
        ("east_right", [[1, 0, 0], [0, 1, 0]]),
        ("north_left", [[0, 1, 0], [1, 0, 0]]),
    ],
)
def test_angle_layout_azimuth_convention(mode_mono, convention, expected):
    # Azimuth conventions are properly applied
    layout = AngleLayout(
        angles=[[90, 0], [90, 90]] * ureg.deg, azimuth_convention=convention
    )
    np.testing.assert_allclose(layout.directions, expected, atol=1e-8)


def test_direction_layout(mode_mono):
    """
    Unit tests for the DirectionLayout class
    """
    # Constructing without argument fails
    with pytest.raises(TypeError):
        DirectionLayout()

    # (3,) arrays are reshaped as (1, 3)
    assert DirectionLayout([0, 0, 1]).directions.shape == (1, 3)

    # Regular construction pattern succeeds
    layout = DirectionLayout(
        [
            [0, 0, 1],
            [np.sqrt(2) / 2, 0, np.sqrt(2) / 2],
            [0, np.sqrt(2) / 2, np.sqrt(2) / 2],
            [-np.sqrt(2) / 2, 0, np.sqrt(2) / 2],
        ],
    )

    # Angles are correctly computed
    np.testing.assert_allclose(
        layout.angles.m_as(ureg.deg),
        [[0, 0], [45, 0], [45, 90], [45, 180]],
    )


@pytest.mark.parametrize(
    "convention, expected",
    [
        ("east_right", [[90, 0], [90, 90]]),
        ("north_left", [[90, 90], [90, 0]]),
    ],
)
def test_direction_layout_azimuth_convention(mode_mono, convention, expected):
    # Azimuth conventions are properly applied
    layout = DirectionLayout(
        directions=[[1, 0, 0], [0, 1, 0]], azimuth_convention=convention
    )
    np.testing.assert_allclose(layout.angles.m_as(ureg.deg), expected, atol=1e-8)


def test_azimuth_ring_layout(mode_mono):
    # Constructing without arguments fails
    with pytest.raises(TypeError):
        AzimuthRingLayout()

    # Regular construction pattern succeeds
    layout = AzimuthRingLayout(
        45 * ureg.deg, np.linspace(0, 2 * np.pi, 8, endpoint=False) * ureg.rad
    )

    # Angles are correctly computed
    np.testing.assert_allclose(
        layout.angles.m_as(ureg.deg),
        [
            [45, 0],
            [45, 45],
            [45, 90],
            [45, 135],
            [45, 180],
            [45, 225],
            [45, 270],
            [45, 315],
        ],
    )

    # We don't test the direction generation (it's the default implementation)


@pytest.mark.parametrize(
    "convention, expected",
    [
        ("east_right", [[1, 0, 0], [0, 1, 0]]),
        ("north_left", [[0, 1, 0], [1, 0, 0]]),
    ],
)
def test_azimuth_ring_layout_azimuth_convention(mode_mono, convention, expected):
    # Azimuth conventions are properly applied
    layout = AzimuthRingLayout(
        90 * ureg.deg,
        [0, 90] * ureg.deg,
        azimuth_convention=convention,
    )
    np.testing.assert_allclose(layout.directions, expected, atol=1e-8)


def test_hemisphere_plane_layout(mode_mono):
    # Constructing without arguments fails
    with pytest.raises(TypeError):
        HemispherePlaneLayout()

    # Regular construction pattern succeeds
    layout = HemispherePlaneLayout(
        np.linspace(-90, 90, 5, endpoint=True) * ureg.deg, 0 * ureg.deg
    )

    # Angles are correctly computed
    assert np.allclose(
        layout.angles,
        [[-90, 0], [-45, 0], [0, 0], [45, 0], [90, 0]] * ureg.deg,
    )

    # We don't test the direction generation (it's the default implementation)


@pytest.mark.parametrize(
    "convention, expected",
    [
        ("east_right", [[-1, 0, 0], [0, 0, 1], [1, 0, 0]]),
        ("north_left", [[0, -1, 0], [0, 0, 1], [0, 1, 0]]),
    ],
)
def test_hemisphere_plane_layout_azimuth_convention(mode_mono, convention, expected):
    # Azimuth conventions are properly applied
    layout = HemispherePlaneLayout(
        [-90, 0, 90] * ureg.deg,
        0 * ureg.deg,
        azimuth_convention=convention,
    )
    np.testing.assert_allclose(layout.directions, expected, atol=1e-8)


def test_grid_layout(mode_mono):
    # Constructing without arguments fails
    with pytest.raises(TypeError):
        GridLayout()

    # Regular construction pattern succeeds
    layout = GridLayout([0, 45] * ureg.deg, [0, 180] * ureg.deg)

    # Angles are correctly computed
    assert np.allclose(
        layout.angles,
        [[0, 0], [45, 0], [0, 180], [45, 180]] * ureg.deg,
    )

    # We don't test the direction generation (it's the default implementation)


@pytest.mark.parametrize(
    "convention, expected",
    [
        ("east_right", [[0, 0, 1], [1, 0, 0], [0, 0, 1], [0, 1, 0]]),
        ("north_left", [[0, 0, 1], [0, 1, 0], [0, 0, 1], [1, 0, 0]]),
    ],
)
def test_grid_layout_azimuth_convention(mode_mono, convention, expected):
    # Azimuth conventions are properly applied
    layout = GridLayout(
        [0, 90] * ureg.deg,
        [0, 90] * ureg.deg,
        azimuth_convention=convention,
    )
    np.testing.assert_allclose(layout.directions, expected, atol=1e-8)


# ------------------------------------------------------------------------------
#                       MultiDistantMeasure implementation
# ------------------------------------------------------------------------------


def test_multi_distant_measure_construct(mode_mono):
    """
    Basic constructor testing for MultiDistantMeasure.
    """
    # Constructing without argument succeeds
    measure = MultiDistantMeasure(direction_layout=[0, 45] * ureg.deg)
    check_scene_element(measure, mi.Sensor)


def test_multi_distant_measure_construct_specific(mode_mono):
    """
    Class method constructor testing for MultiDistantMeasure
    """
    # Unitless values are attached default units
    measure = MultiDistantMeasure.hplane(zeniths=[0, 45], azimuth=0)
    assert isinstance(measure.direction_layout, HemispherePlaneLayout)
    assert np.allclose(measure.direction_layout.angles, [[0, 0], [45, 0]] * ureg.deg)

    # The dictionary conversion protocol applies units, including when used with
    # a class method constructor
    measure = measure_factory.convert(
        {
            "type": "mdistant",
            "construct": "from_angles",
            "angles": [np.pi / 4, 0],
            "angles_units": "rad",
        }
    )
    assert isinstance(measure.direction_layout, AngleLayout)
    assert np.allclose(measure.direction_layout.angles, [[45, 0]] * ureg.deg)


def test_multi_distant_measure_viewing_angles(mode_mono):
    """
    Unit tests for :attr:`.MultiDistantMeasure.viewing_angles`.
    """
    # Viewing angle computation is correct
    measure = MultiDistantMeasure(
        direction_layout=[
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ]
    )

    expected = (
        np.array(
            [
                [0, 0],
                [45, 0],
                [45, 90],
                [np.rad2deg(np.arccos(1 / np.sqrt(3))), 45],
            ]
        ).reshape((4, 1, 2))
        * ureg.deg
    )

    assert np.allclose(measure.viewing_angles, expected)

    # Directions which would normally map to the [-π, 0] domain are normalised
    # to [0, 2π]
    measure = MultiDistantMeasure(direction_layout=[[0, -1, 1]])
    expected = [45, 270] * ureg.deg
    assert np.allclose(measure.viewing_angles, expected)



import os
import tempfile

import numpy as np
import pytest

from eradiate.scenes.biosphere import AbstractTree, InstancedCanopyElement, LeafCloud
from eradiate.scenes.core import traverse
from eradiate.test_tools.types import check_scene_element

# ------------------------------------------------------------------------------
#                            Fixture definitions
# ------------------------------------------------------------------------------


@pytest.fixture(scope="module")
def tempfile_spheres():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_leaves.txt")
        with open(filename, "w") as tf:
            tf.write("8.864 9.040 1.878\n")
            tf.write("0.100 9.539 -10.463\n")
            tf.write("0.100 -2.274 -9.204\n")
            tf.write("0.100 -9.957 -4.971\n")
            tf.write("0.100 5.339 9.153\n")
        yield filename


# ------------------------------------------------------------------------------
#                        InstancedCanopyElement Tests
# ------------------------------------------------------------------------------


def test_instanced_canopy_element_create(mode_mono):
    """Unit testing for the :class:`InstancedCanopyElement` constructor."""
    cloud = LeafCloud(
        leaf_positions=[[0, 0, 0]], leaf_orientations=[[0, 0, 1]], leaf_radii=[0.1]
    )
    tree = AbstractTree(
        leaf_cloud=cloud, trunk_height=1.0, trunk_radius=0.1, trunk_reflectance=0.5
    )
    positions = np.array([[-5, -5, -5], [5, 5, 5]])
    assert InstancedCanopyElement(canopy_element=cloud, instance_positions=positions)
    assert InstancedCanopyElement(canopy_element=tree, instance_positions=positions)


def test_instanced_canopy_element_kernel_dict(mode_mono):
    """Unit testing for :meth:`InstancedLeafCloud.kernel_dict`."""
    cloud = LeafCloud(
        leaf_positions=[[0, 0, 0]],
        leaf_orientations=[[0, 0, 1]],
        leaf_radii=[0.1],
        id="leaf_cloud",
    )
    tree = AbstractTree(
        leaf_cloud=cloud,
        trunk_height=1.0,
        trunk_radius=0.1,
        trunk_reflectance=0.5,
        id="tree",
    )
    positions = np.array([[-5, -5, -5], [5, 5, 5]])
    instances = InstancedCanopyElement(
        canopy_element=tree, instance_positions=positions, id="element"
    )

    # The generated kernel dictionary can be instantiated
    check_scene_element(instances)
    template, params = traverse(instances)
    assert set(params.keys()) == {
        "bsdf_leaf_cloud.reflectance.value",
        "bsdf_leaf_cloud.transmittance.value",
        "bsdf_tree.reflectance.value",
    }


def test_instanced_leaf_cloud_from_file(mode_mono, tempfile_spheres):
    """Unit testing for :meth:`InstancedLeafCloud.from_file`."""
    assert InstancedCanopyElement.from_file(
        filename=tempfile_spheres,
        canopy_element={
            "type": "leaf_cloud",
            "leaf_positions": [[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]],
            "leaf_orientations": [[0, 0, 1], [1, 0, 0]],
            "leaf_radii": [0.1, 0.1],
        },
    )



import mitsuba as mi
import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.exceptions import UnsupportedModeError
from eradiate.experiments import CanopyExperiment
from eradiate.scenes.biosphere import DiscreteCanopy
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.scenes.spectra import MultiDeltaSpectrum
from eradiate.test_tools.types import check_scene_element


def test_canopy_experiment_construct_default(modes_all_double):
    """
    CanopyExperiment initializes with default params in all modes
    """
    assert CanopyExperiment()


def test_canopy_experiment_construct_measures(mode_mono_double):
    """
    A variety of measure specifications are acceptable
    """

    # Init with a single measure (not wrapped in a sequence)
    assert CanopyExperiment(measures=MultiDistantMeasure())

    # Init from a dict-based measure spec
    # -- Correctly wrapped in a sequence
    assert CanopyExperiment(measures=[{"type": "distant"}])
    # -- Not wrapped in a sequence
    assert CanopyExperiment(measures={"type": "distant"})


@pytest.mark.parametrize("padding", (0, 1))
def test_canopy_experiment_construct_normalize_measures(mode_mono_double, padding):
    """
    When canopy is not None, measure target matches canopy unit cell
    """
    exp = CanopyExperiment(
        canopy=DiscreteCanopy.homogeneous(
            lai=3.0,
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
            padding=padding,
        ),
        measures=MultiDistantMeasure(),
    )
    target = exp.measures[0].target
    canopy = exp.canopy
    assert np.isclose(target.xmin, -0.5 * canopy.size[0])
    assert np.isclose(target.xmax, 0.5 * canopy.size[0])
    assert np.isclose(target.ymin, -0.5 * canopy.size[1])
    assert np.isclose(target.ymax, 0.5 * canopy.size[1])


@pytest.mark.parametrize("padding", (0, 1))
def test_canopy_experiment_kernel_dict(modes_all_double, padding):
    # Surface width is appropriately inherited from canopy
    exp = CanopyExperiment(
        canopy=DiscreteCanopy.homogeneous(
            lai=0.5,  # This very low value ensures fast object initialization
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
            padding=padding,
        )
    )
    mi_wrapper = check_scene_element(
        exp.scene, mi.Scene, drop_parameters=False
    )  # Don't drop untracked params at this stage: we want to check the surface transform

    assert np.allclose(
        mi_wrapper.parameters["surface_shape.to_world"].transform_affine(
            mi.Point3f(1, -1, 0)
        ),
        [5 * (2 * padding + 1), -5 * (2 * padding + 1), 0],
    )


@pytest.mark.slow
def test_canopy_experiment_real_life(modes_all_double):
    # Construct with typical parameters
    exp = CanopyExperiment(
        surface={"type": "lambertian"},
        canopy={
            "type": "discrete_canopy",
            "construct": "homogeneous",
            "lai": 3.0,
            "leaf_radius": 0.1 * ureg.m,
            "l_horizontal": 10.0 * ureg.m,
            "l_vertical": 2.0 * ureg.m,
        },
        illumination={"type": "directional", "zenith": 45.0},
        measures={
            "type": "distant",
            "construct": "hplane",
            "zeniths": np.arange(-60, 61, 5),
            "azimuth": 0.0,
        },
    )
    check_scene_element(exp.scene, mi.Scene)


@pytest.mark.slow
def test_canopy_experiment_run_detailed(modes_all_double):
    """
    Test for correctness of the result dataset generated by CanopyExperiment.
    Note: This test is outdated, most of its content should be transferred to
    tests for measure post-processing pipelines.
    """
    srf = MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm)
    if eradiate.mode().is_mono:
        expected_vars = {"brdf", "brf", "irradiance", "radiance"}
        expected_coords_radiance = {
            "saa",
            "sza",
            "vaa",
            "vza",
            "w",
            "x_index",
            "x",
            "y_index",
            "y",
        }
        expected_coords_irradiance = {"saa", "sza", "w"}

    elif eradiate.mode().is_ckd:
        expected_vars = {"brdf", "brf", "irradiance", "radiance"}
        expected_coords_radiance = {
            "bin_wmin",
            "bin_wmax",
            "saa",
            "sza",
            "vaa",
            "vza",
            "w",
            "x_index",
            "x",
            "y_index",
            "y",
        }
        expected_coords_irradiance = {"bin_wmin", "bin_wmax", "saa", "sza", "w"}

    else:
        raise UnsupportedModeError

    exp = CanopyExperiment(
        measures=[
            {
                "type": "hemispherical_distant",
                "id": "toa_hsphere",
                "film_resolution": (32, 32),
                "spp": 1000,
                "srf": srf,
            },
        ]
    )

    results = eradiate.run(exp)

    # Post-processing creates expected variables ...
    assert set(results.data_vars) == expected_vars

    # ... dimensions
    assert set(results["radiance"].dims) == {"saa", "sza", "w", "x_index", "y_index"}
    assert set(results["irradiance"].dims) == {"saa", "sza", "w"}

    # ... and other coordinates
    assert set(results["radiance"].coords) == expected_coords_radiance
    assert set(results["irradiance"].coords) == expected_coords_irradiance

    # We just check that we record something as expected
    assert np.all(results["radiance"].data > 0.0)



import pytest

import eradiate
from eradiate import unit_registry as ureg


@pytest.mark.parametrize("geometry", ["plane_parallel", "spherical_shell"])
def test_heterogeneous_parameter_lookup(
    modes_all_double, geometry, atmosphere_us_standard_mono, atmosphere_us_standard_ckd
):
    """
    Kernel parameter lookup
    =======================

    This test checks if parameters belonging to objects referenced by Mitsuba
    ``Medium`` objects are correctly retrieved.

    Rationale
    ---------

    Parameter paths are generated upon Mitsuba scene traversal based on the
    order in which they are discovered. The traversal order, although
    deterministic, is not easy to predict. Consequently, Eradiate implements a
    parameter lookup system which seeks a parameter during Mitsuba scene
    traversal based on its ID. This allows to safely retrieve Mitsuba parameter
    paths, regardless the number of objects which can register them as children
    or their discovery order.

    For that purpose, we create an ``Experiment`` and populate it with an
    atmosphere and an *in situ* measure. This way, the ``Medium`` instance
    generated by the atmosphere configuration is referenced both by the
    atmosphere shape and the sensor.

    Expected behaviour
    ------------------

    The sensor is discovered first because of the ID it is assigned (``dflux``).
    Consequently, although the medium is emitted upon atmosphere processing, the
    parameter associated with the volume data buffer nested in the ``Medium`` has
    the sensor in its parent nodes rather than the atmosphere shape.
    """
    atmosphere_mono = atmosphere_us_standard_mono
    atmosphere_mono.update({"has_scattering": False})

    atmosphere_ckd = atmosphere_us_standard_ckd
    atmosphere_ckd.update({"has_scattering": False})

    exp = eradiate.experiments.AtmosphereExperiment(
        geometry=geometry,
        surface={"type": "lambertian", "reflectance": 1.0},
        atmosphere=atmosphere_ckd if eradiate.mode().is_ckd else atmosphere_mono,
        illumination={"type": "directional", "zenith": 30.0},
        measures={
            "type": "distant_flux",
            "id": "dflux",
            "srf": {
                "type": "multi_delta",
                "wavelengths": 550.0 * ureg.nm,
            },
            "ray_offset": 1.0 * ureg.m,
        },
    )
    exp.init()

    # Medium is resolved, regardless the fact that it is first encountered as
    # a member of the "dflux" sensor
    assert (
        exp.mi_scene.umap_template["medium_atmosphere.albedo.volume.data"].parameter_id
        == "dflux.medium.albedo.volume.data"
        if geometry == "spherical_shell"
        else "dflux.medium.albedo.data"
    )



"""Test cases with AtmosphereExperiment and an RPV surface."""

import os

import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate import unit_registry as ureg


def map_to_0_360(x: float) -> float:
    return x % 360


@pytest.mark.parametrize("illumination_azimuth", [0.0, 30.0, 120.0, 210.0, 300.0])
def test_film_to_angular_coord_conversion_multi_distant(
    mode_mono, illumination_azimuth, artefact_dir
):
    r"""
    Film to angular coordinates conversion (``multi_distant``)
    ==========================================================

    This test case asserts the correct conversion from film coordinates (x, y)
    to angular coordinates (theta, phi).

    We check that the conversion is correct by computing the BRF of a scene
    built such that it is forward/backward scattering and we verify that the
    BRF is consistent with this scattering behaviour.


    Rationale
    ---------

    * Geometry: a square surface with unit size and a RPV BRDF with
      reflectance :math:`\rho_0 = 0.1`.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta = 30.0°` and an azimuth angle
      :math:`\varphi \in [0.0, 30.0, 120.0, 210.0, 300.0]°`.
    * Sensor: Distant reflectance measure covering a plane (20 angular points,
      1 sample per pixel).


    Expected behaviour
    ------------------

    Within the scope of this test, we define the forward region of the
    principal plane by:

    .. math:: \varphi_v = \varphi_i + 180°
       :label: multi_distant_forward

    where:

    * :math:`\varphi_v` is the viewing azimuth angle [deg]
    * :math:`\varphi_i` is the illumination azimuth angle [deg]

    Correspondingly, the backward region of the principal plane is defined by:

    .. math:: \varphi_v = \varphi_i
       :label: multi_distant_backward

    We expect that the mean BRF value will be larger/smaller in the forward
    region :eq:`multi_distant_forward` than in the backward region
    :eq:`multi_distant_backward` for a forward/backward scattering scene,
    respectively.

    Results
    -------

    .. image:: generated/plots/test_ftacc_multi_distant_forward_0.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_forward_30.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_forward_120.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_forward_210.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_forward_300.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_backward_0.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_backward_30.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_backward_120.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_backward_210.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_multi_distant_backward_300.0.png
       :width: 95%
    """

    spp = 1
    reflectance = 0.1
    n_vza = 20
    measure_azimuth = illumination_azimuth
    gs = {"forward": 0.1, "backward": -0.1}

    experiments = {
        orientation: eradiate.experiments.AtmosphereExperiment(
            measures={
                "type": "mdistant",
                "construct": "hplane",
                "azimuth": illumination_azimuth,
                "zeniths": np.linspace(-89, 89, n_vza) * ureg.deg,
                "spp": spp,
            },
            illumination={
                "type": "directional",
                "zenith": 30.0 * ureg.deg,
                "azimuth": illumination_azimuth,
            },
            atmosphere=None,
            surface={
                "type": "rpv",
                "g": g,
                "rho_0": reflectance,
                "k": 0.9,
            },
        )
        for orientation, g in gs.items()
    }

    results = {name: eradiate.run(exp) for name, exp in experiments.items()}

    def select_brf(brf, orientation):
        relative_azimuth = map_to_0_360(measure_azimuth - illumination_azimuth)

        if relative_azimuth == 0.0:
            return (
                brf.where(brf.vza < 0.0)
                if orientation == "forward"
                else brf.where(brf.vza > 0.0)
            )
        elif relative_azimuth == 180.0:
            return (
                brf.where(brf.vza > 0.0)
                if orientation == "forward"
                else brf.where(brf.vza < 0.0)
            )
        else:
            raise ValueError(
                "cannot compute forward BRF when the relative azimuth is not "
                "either 0 or 180."
            )

    def is_forward_scattering(brf):
        forward_brf = select_brf(brf, "forward")
        backward_brf = select_brf(brf, "backward")
        return forward_brf.mean().values > backward_brf.mean().values

    def make_figure(brf, g, artefact_dir, orientation: str):
        brf_forward = select_brf(brf, "forward")
        brf_backward = select_brf(brf, "backward")

        brf_values = np.concatenate([brf_forward.values, brf_backward.values])
        brf_max = np.nanmax(brf_values)
        brf_min = np.nanmin(brf_values)
        ylim = [0.9 * brf_min, 1.1 * brf_max]

        desc = orientation.title()

        fig = plt.figure(figsize=(8, 3))
        plt.suptitle(
            f"{desc} scattering RPV surface (g={g}), "
            f"illumination azimuth = {illumination_azimuth}°"
        )

        with plt.rc_context({"lines.linestyle": ":", "lines.marker": "."}):
            for orientation, brf, sub, ticks, xtext in zip(
                ["forward", "backward"],
                [brf_forward, brf_backward],
                [(1, 2, 1), (1, 2, 2)],
                [[-90, -60, -30, 0], [0, 30, 60, 90]],
                [-45, 45],
            ):
                plt.subplot(*sub)
                ax = plt.gca()
                ax.set_xticks(ticks)
                ax.set_xticklabels(list(map(str, ticks)))
                brf.plot(x="vza", xlim=[-95, 0], ylim=ylim)
                mean = float(brf.mean().values)
                plt.text(
                    s=f"mean = {mean:.2e}",
                    x=xtext,
                    y=(brf_max + brf_min) / 2.0,
                    ha="center",
                    color="red",
                )
                plt.title(f"{orientation} BRF")

        filename = f"test_ftacc_multi_distant_{desc.lower()}_{illumination_azimuth}.png"
        outdir = os.path.join(artefact_dir, "plots")
        os.makedirs(outdir, exist_ok=True)
        fname_plot = os.path.join(outdir, filename)
        plt.tight_layout()
        fig.savefig(fname_plot, dpi=200)
        plt.close()

    for orientation, result in results.items():
        make_figure(result.brf, gs[orientation], artefact_dir, orientation)

    assert is_forward_scattering(results["forward"].brf)
    assert not is_forward_scattering(brf=results["backward"].brf)


def where_azimuth(
    da: xr.DataArray, name: str, start: float, stop: float
) -> xr.DataArray:
    """
    Select data for viewing azimuth in ['start', 'stop']°.
    """
    if start < stop:
        return da.where((da.vaa > start) & (da.vaa < stop))
    else:
        return xr.merge(
            [
                da.where((da.vaa > start) & (da.vaa < 360.0)),
                da.where((da.vaa > 0.0) & (da.vaa < stop)),
            ]
        )[name]


def select_orientation(
    da: xr.DataArray,
    name: str,
    illumination_azimuth: float,
    orientation: str,
) -> xr.DataArray:
    r"""
    Select hemispherical data in requested orientation.

    If :math:`\varphi_i` is the illumination azimuth and :math:`\varphi_v` is the
    viewing azimuth, the forward region is defined by:

    .. math::

        \varphi_i + 90° \ll \varphi_v \ll \varphi_i + 270°

    Parameters
    ----------
    da: DataArray
        Sector radiosity at an infinite distance in the hemisphere.

    name: str
        Name of the data variable to select.

    illumination_azimuth: float
        Azimuth [deg].

    orientation : {"forward", "backward"}
        Selected orientation.

    Returns
    -------
    DataArray
        Data in forward direction.
    """
    return (
        where_azimuth(
            da=da,
            name=name,
            start=map_to_0_360(illumination_azimuth + 90.0),
            stop=map_to_0_360(illumination_azimuth + 270.0),
        )
        if orientation == "forward"
        else where_azimuth(
            da=da,
            name=name,
            start=map_to_0_360(illumination_azimuth + 270.0),
            stop=map_to_0_360(illumination_azimuth + 90.0),
        )
    )


def is_forward_scattering(
    da: xr.DataArray, name: str, illumination_azimuth: float
) -> bool:
    """
    Determines if a scene is forward scattering.

    Parameters
    ----------
    da: DataArray
        Sector radiosity at an infinite distance in the hemisphere.

    name: str
        Name of the hemispherical variable.

    illumination_azimuth: float
        Azimuth [deg].
    """
    da_forward = select_orientation(da, name, illumination_azimuth, "forward")
    da_backward = select_orientation(da, name, illumination_azimuth, "backward")
    return da_forward.mean().values > da_backward.mean().values


def make_figure(
    results: xr.Dataset,
    name: str,
    g: float,
    forward: bool,
    illumination_azimuth: float,
    measure: str,
    res: int,
    artefact_dir,
):
    fig = plt.figure(figsize=(8, 3))

    desc = "Forward" if forward else "Backward"
    plt.suptitle(
        f"{desc} scattering RPV surface (g={g}), "
        f"illumination azimuth = {illumination_azimuth}°"
    )

    for orientation, sub in zip(
        ["forward", "backward"],
        [(1, 2, 1), (1, 2, 2)],
    ):
        plt.subplot(*sub)
        ax = plt.gca()
        ax.set_aspect("equal")
        da = select_orientation(
            results[name],
            name,
            illumination_azimuth,
            "forward",
        )
        da.plot()

        mean = float(da.mean().values)
        plt.text(
            s=f"mean = {mean:.2e}",
            x=res / 2,
            y=res / 2,
            ha="center",
            color="red",
        )
        plt.title(f"{orientation} {name}")

    filename = f"test_ftacc_{measure}_{desc.lower()}_{illumination_azimuth}.png"
    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    fname_plot = os.path.join(outdir, filename)
    plt.tight_layout()
    fig.savefig(fname_plot, dpi=200)
    plt.close()


@pytest.mark.parametrize("illumination_azimuth", [0.0, 30.0, 120.0, 210.0, 300.0])
def test_film_to_angular_coord_conversion_distant_flux(
    mode_mono, illumination_azimuth, artefact_dir
):
    r"""
    Film to angular coordinates conversion (``distant_flux``)
    =========================================================

    This test case asserts the correct conversion from film coordinates (x, y)
    to angular coordinates (theta, phi) with the ``distant_flux`` measure.

    Since the mapping is only visible in post-processed results, this test
    requires the Experiment framework.

    We setup two experiments such that the first one includes a forward
    scattering scene and the second one a backward scattering scene.

    This is achieved using a surface reflectance model (``RPV``) that provides a
    parameter (``g``) to control the forward/backward scattering behaviour of the
    surface.

    We set a ``distant_flux`` measure to compute the sector radiosity at infinite
    distance in the hemisphere whose direction is the upward direction.

    Rationale
    ---------

    * Geometry: a square surface with unit size and a RPV BRDF with
      reflectance :math:`\rho_0 = 0.1` and with asymmetry parameter :math:`g`
      set to 0.1 and -0.1 for the forward and backward scattering scenes,
      respectively.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta_i = 30.0°` and an azimuth angle
      :math:`\varphi_i \in [0.0, 30.0, 120.0, 210.0, 300.0]°`.
    * Sensor: Distant flux measure covering a hemisphere (32 * 32 angular
      points, 1 sample per pixel).


    Expected behaviour
    ------------------

    If :math:`\varphi_i` is the illumination azimuth and :math:`\varphi_v` is the
    viewing azimuth, we define the forward region of the hemisphere by:

    .. math:: \phi_i + 90° < \phi_v < \phi_i + 270°
       :label: distant_flux_forward

    and correspondingly, the backward region, by:

    .. math:: \phi_i + 270° < \phi_v < \phi_i + 90°
       :label: distant_flux_backward

    We then compute their mean value within the azimuth angle range where they
    are non zero.

    We expect that the forward scattering scene will yield to a measured sector
    radiosity being larger in the range of azimuth angles given by
    :eq:`distant_flux_forward` than in the range of azimuth angles given by
    :eq:`distant_flux_backward`.
    We expect the opposite result for the backward scattering scene.

    Results
    -------

    .. image:: generated/plots/test_ftacc_distant_flux_forward_0.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_forward_30.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_forward_120.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_forward_210.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_forward_300.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_backward_0.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_backward_30.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_backward_120.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_backward_210.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_distant_flux_backward_300.0.png
       :width: 95%
    """

    spp = 1
    res = 32
    reflectance = 0.1
    gs = {"forward": 0.1, "backward": -0.1}

    experiments = {
        orientation: eradiate.experiments.AtmosphereExperiment(
            measures={
                "type": "distant_flux",
                "direction": [0, 0, 1],
                "film_resolution": (res, res),
                "spp": spp,
            },
            illumination={
                "type": "directional",
                "zenith": 30.0 * ureg.deg,
                "azimuth": illumination_azimuth,
            },
            atmosphere=None,
            surface={"type": "rpv", "g": g, "rho_0": reflectance, "k": 0.9},
        )
        for orientation, g in gs.items()
    }

    results = {
        orientation: eradiate.run(exp) for orientation, exp in experiments.items()
    }

    for orientation, g in gs.items():
        make_figure(
            results=results[orientation],
            name="sector_radiosity",
            g=g,
            forward=(orientation == "forward"),
            illumination_azimuth=illumination_azimuth,
            measure="distant_flux",
            res=res,
            artefact_dir=artefact_dir,
        )

    assert is_forward_scattering(
        da=results["forward"].sector_radiosity,
        name="sector_radiosity",
        illumination_azimuth=illumination_azimuth,
    )
    assert not is_forward_scattering(
        da=results["backward"].sector_radiosity,
        name="sector_radiosity",
        illumination_azimuth=illumination_azimuth,
    )


@pytest.mark.parametrize("illumination_azimuth", [0.0, 30.0, 120.0, 210.0, 300.0])
def test_film_to_angular_coord_conversion_hemispherical_distant(
    mode_mono_double, illumination_azimuth, artefact_dir
):
    r"""
    Film to angular coordinates conversion (``hdistant``)
    =====================================================

    This test case asserts the correct conversion from film coordinates (x, y)
    to angular coordinates (theta, phi) with the ``hdistant`` measure.

    Since the mapping is only visible in post-processed results, this test
    requires the Experiment framework.

    We setup two experiments such that the first one includes a forward
    scattering scene and the second one a backward scattering scene.

    This is achieved using a surface reflectance model (``RPV``) that provides a
    parameter (``g``) to control the forward/backward scattering behaviour of the
    surface.

    We set a ``hemispherical_distant`` measure to compute the BRF at infinite
    distance in the hemisphere whose direction is the upward direction.

    Rationale
    ---------

    * Geometry: a square surface with unit size and a RPV BRDF with
      reflectance :math:`\rho_0 = 0.1` and with asymmetry parameter :math:`g`
      set to 0.1 and -0.1 for the forward and backward scattering scenes,
      respectively.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta_i = 30.0°` and an azimuth angle
      :math:`\varphi_i \in [0.0, 30.0, 120.0, 210.0, 300.0]°`.
    * Sensor: Distant hemispherical measure covering the hemisphere (32 * 32
      angular points, 1 sample per pixel).


    Expected behaviour
    ------------------

    If :math:`\varphi_i` is the illumination azimuth and :math:`\varphi_v` is the
    viewing azimuth, we define the forward region of the hemisphere by:

    .. math:: \phi_i + 90° < \phi_v < \phi_i + 270°
       :label: hemispherical_distant_forward

    and correspondingly, the backward region, by:

    .. math:: \phi_i + 270° < \phi_v < \phi_i + 90°
       :label: hemispherical_distant_backward

    We then compute their mean value within the azimuth angle range where they
    are non zero.

    We expect that the forward scattering scene will yield to a measured BRF
    being larger in the range of azimuth angles given by
    :eq:`hemispherical_distant_forward` than in the range of azimuth angles
    given by :eq:`hemispherical_distant_backward`.
    We expect the opposite result for the backward scattering scene.

    Results
    -------

    .. image:: generated/plots/test_ftacc_hdistant_forward_0.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_forward_30.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_forward_120.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_forward_210.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_forward_300.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_backward_0.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_backward_30.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_backward_120.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_backward_210.0.png
       :width: 95%
    .. image:: generated/plots/test_ftacc_hdistant_backward_300.0.png
       :width: 95%
    """

    spp = 1
    res = 32
    reflectance = 0.1
    gs = {"forward": 0.1, "backward": -0.1}

    experiments = {
        orientation: eradiate.experiments.AtmosphereExperiment(
            measures={
                "type": "hdistant",
                "direction": [0, 0, 1],
                "film_resolution": (res, res),
                "spp": spp,
            },
            illumination={
                "type": "directional",
                "zenith": 30.0 * ureg.deg,
                "azimuth": illumination_azimuth,
            },
            atmosphere=None,
            surface={"type": "rpv", "g": g, "rho_0": reflectance, "k": 0.9},
        )
        for orientation, g in gs.items()
    }

    results = {
        orientation: eradiate.run(exp) for orientation, exp in experiments.items()
    }

    for orientation, g in gs.items():
        make_figure(
            results=results[orientation],
            name="brf",
            g=g,
            forward=(orientation == "forward"),
            illumination_azimuth=illumination_azimuth,
            measure="hdistant",
            res=res,
            artefact_dir=artefact_dir,
        )

    assert is_forward_scattering(
        da=results["forward"].brf,
        name="brf",
        illumination_azimuth=illumination_azimuth,
    )
    assert not is_forward_scattering(
        da=results["backward"].brf,
        name="brf",
        illumination_azimuth=illumination_azimuth,
    )


@pytest.mark.slow
@pytest.mark.parametrize(
    "atmosphere", [None, "homogeneous"], ids=["none", "homogeneous"]
)
@pytest.mark.parametrize("reflectance", [0.0, 0.5, 1.0])
def test_rpv_vs_lambertian(
    ert_seed_state, mode_mono, atmosphere, reflectance, artefact_dir, request
):
    r"""
    RPV(:math:`\rho, g=0, k=1, rho_c=1`) equivalent to Lambertian(:math:`\rho`)
    ===========================================================================

    A surface with a lambertian BSDF behaves like the same surface but with a
    RPV BSDF with identical reflectance value, :math:`g = 0`, :math:`k = 1`
    and :math:`\rho_c = 1`.

    Rationale
    ---------

    * Geometry: a square surface with unit size and

      * experiment 1: a lambertian BSDF with reflectance :math:`\rho_0 = 1.0`.
      * experiment 2: a RPV BSDF with :math:`\rho_0 = 1.0`, :math:`g = 0`,
        :math:`k = 1`, :math:`rho_c = 1`.

    * Atmosphere: No atmosphere or default homogeneous atmosphere.
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta = 30.0°` and an azimuth angle :math:`\varphi = 0.0`.
    * Sensor: Distant reflectance measure covering a plane (11 angular points,
      1 sample per pixel with no atmosphere, 10000 with atmosphere).

    Expected behaviour
    ------------------

    Experiment 1 and 2 TOA BRF values are equal.

    Results
    -------

    .. image:: generated/plots/test_rpv_vs_lambertian-none-0.0.png
       :width: 95%
    .. image:: generated/plots/test_rpv_vs_lambertian-none-0.5.png
       :width: 95%
    .. image:: generated/plots/test_rpv_vs_lambertian-none-1.0.png
       :width: 95%

    .. image:: generated/plots/test_rpv_vs_lambertian-homogeneous-0.0.png
       :width: 95%
    .. image:: generated/plots/test_rpv_vs_lambertian-homogeneous-0.5.png
       :width: 95%
    .. image:: generated/plots/test_rpv_vs_lambertian-homogeneous-1.0.png
       :width: 95%
    """

    def make_figure_rpv_vs_lambertian(fname_plot, results: dict, title=""):
        fig = plt.figure(figsize=(8, 3))
        results_lambertian = results["lambertian"]
        results_rpv = results["rpv"]

        with plt.rc_context({"lines.linestyle": "dashed"}):
            results_lambertian.brf.plot(x="vza", marker="o", label="lambertian")
            results_rpv.brf.plot(x="vza", marker="^", label="rpv")

        bias = np.zeros_like(results_lambertian.brf.values)
        np.divide(
            results_rpv.brf.values - results_lambertian.brf.values,
            results_lambertian.brf.values,
            where=results_lambertian.brf.values != 0,
            out=bias,
        )
        mean_bias = np.mean(np.abs(bias))
        plt.annotate(
            text=f"Mean bias: {round(100 * mean_bias, 2)} %",
            xy=(0.5, 0.5),
            horizontalalignment="center",
            xycoords="axes fraction",
        )
        plt.title(title)
        plt.tight_layout()

        print(f"Saving figure to {fname_plot}")
        fig.savefig(fname_plot, dpi=200)
        plt.close()

    bsdfs = {
        "lambertian": {"type": "lambertian", "reflectance": reflectance},
        "rpv": {
            "type": "rpv",
            **dict(
                rho_0=reflectance,
                g=0.0,
                k=1.0,
                rho_c=1.0,
            ),
        },
    }

    experiments = {
        bsdf: eradiate.experiments.AtmosphereExperiment(
            illumination={"type": "directional", "zenith": 30.0 * ureg.deg},
            measures={
                "type": "mdistant",
                "construct": "hplane",
                "zeniths": np.arange(-75, 75, 11),
                "azimuth": 0.0 * ureg.deg,
                "spp": 1 if atmosphere is None else 100000,
            },
            atmosphere=None if atmosphere is None else {"type": atmosphere},
            surface=bsdfs[bsdf],
        )
        for bsdf in bsdfs.keys()
    }

    # Run experiments
    results = {}
    for bsdf, exp in experiments.items():
        ert_seed_state.reset()
        results[bsdf] = eradiate.run(exp)

    # Make figure
    filename = f"{request.node.originalname}-{'none' if atmosphere is None else 'homogeneous'}-{reflectance}.png"
    outdir = os.path.join(artefact_dir, "plots")
    os.makedirs(outdir, exist_ok=True)
    fname_plot = os.path.join(outdir, filename)
    title = (
        f"{reflectance=}, no atmosphere"
        if atmosphere is None
        else f"{reflectance=}, homogeneous atmosphere"
    )
    make_figure_rpv_vs_lambertian(fname_plot, results, title=title)

    # assert result BRF values are equal
    lambertian = np.squeeze(results["lambertian"].brf.values)
    rpv = np.squeeze(results["rpv"].brf.values)

    if atmosphere is None:
        np.testing.assert_array_equal(rpv, lambertian)
    else:
        np.testing.assert_allclose(rpv, lambertian, rtol=1e-2, atol=1e-2)



import mitsuba as mi
import pytest

from eradiate.scenes.bsdfs import RPVBSDF
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"rho_0": 0.3, "k": 1.4, "g": -0.23},
        {
            "rho_0": 0.3,
            "k": {"type": "uniform", "value": 0.3},
            "g": {
                "type": "interpolated",
                "wavelengths": [300.0, 800.0],
                "values": [-0.23, 0.23],
            },
            "rho_c": 0.2,
        },
    ],
    ids=["noargs", "uniform", "mixed"],
)
def test_rpv_construct(modes_all, kwargs):
    # Default constructor
    assert RPVBSDF(**kwargs)


def test_rpv(modes_all_double):
    bsdf = RPVBSDF(
        rho_0=0.3,
        k={"type": "uniform", "value": 0.3},
        g={
            "type": "interpolated",
            "wavelengths": [300.0, 800.0],
            "values": [-0.23, 0.23],
        },
        rho_c=0.2,
    )

    check_scene_element(bsdf, mi.BSDF)



import numpy as np
import pytest

import eradiate
from eradiate.experiments import AtmosphereExperiment
from eradiate.test_tools.regression import Chi2Test
from eradiate.units import unit_registry as ureg


@pytest.mark.regression
def test_rpv_afgl1986_brfpp(
    mode_ckd_double,
    artefact_dir,
    session_timestamp,
    absorption_database_error_handler_config,
):
    r"""
    RPV AFGL1986 regression test
    ====================================

    This is a regression test, which compares the simulation results of the
    current branch to an older reference version.

    Rationale
    ---------

    This test case uses a basic atmospheric scene:

    * RPV surface emulating a canopy
    * Molecular atmosphere following the AFGL 1986 model

    Parameters

    * Atmosphere: Molecular atmosphere using the agfl1986 profile
    * Surface: Square surface with RPV BSDF with :math:`k = 0.95`, :math:`g = -0.1`
      and :math:`\rho_0 = 0.027685`
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points,
      10000 samples per pixel)

    Expected behaviour
    ------------------

    This test uses the Chi-squared criterion with a threshold of 0.05.

    """
    exp = AtmosphereExperiment(
        surface={"type": "rpv", "k": 0.95, "g": -0.1, "rho_0": 0.027685},
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "irradiance": 20.0,
        },
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "azimuth": 0.0,
                "zeniths": np.arange(-75.0, 75.01, 2.0),
                "srf": {"type": "multi_delta", "wavelengths": 550 * ureg.nm},
            }
        ],
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": {
                "type": "molecular",
                "thermoprops": {
                    "identifier": "afgl_1986-us_standard",
                    "z": np.linspace(0, 120, 61) * ureg.km,
                    "additional_molecules": False,
                },
                "absorption_data": "monotropa",
                "error_handler_config": absorption_database_error_handler_config,
            },
        },
    )
    result = eradiate.run(exp, spp=10000)

    test = Chi2Test(
        name=f"{session_timestamp:%Y%m%d-%H%M%S}-rpv_afgl1986.nc",
        value=result,
        reference="tests/regression_test_references/rpv_afgl1986_brfpp_ref.nc",
        threshold=0.05,
        archive_dir=artefact_dir,
    )

    assert test.run()



import pytest

from eradiate.contexts import MultiGenerator


@pytest.mark.parametrize(
    "input, expected",
    [
        ([range(5), range(10)], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
        ([[0, 1, 2, 3, 0, 2, 1], [0, 1, 2, 3, 0, 2, 1, 4]], [0, 1, 2, 3, 4]),
        ([[3, 0, 2, 1], [0, 4, 2, 0, 1, 4]], [3, 0, 2, 1, 4]),
    ],
)
def test_multi_generator(input, expected):
    assert list(MultiGenerator(input)) == expected



import mitsuba as mi
import numpy as np
import pint
import pytest
from pinttr.exceptions import UnitsError

import eradiate
from eradiate import unit_context_config as ucc
from eradiate import unit_context_kernel as uck
from eradiate import unit_registry as ureg
from eradiate.scenes.spectra import InterpolatedSpectrum, spectrum_factory
from eradiate.spectral import Bin, BinSet
from eradiate.spectral.index import SpectralIndex
from eradiate.test_tools.types import check_scene_element
from eradiate.units import PhysicalQuantity


@pytest.mark.parametrize(
    "tested, expected, units",
    [
        ({}, TypeError, None),
        (
            {"wavelengths": [500.0, 600.0]},
            TypeError,
            None,
        ),
        (
            {"values": [0.0, 1.0]},
            TypeError,
            None,
        ),
        (
            {"wavelengths": [500.0, 600.0], "values": [0.0, 1.0, 2.0]},
            ValueError,
            None,
        ),
        (
            {"wavelengths": [500.0, 600.0], "values": [0.0, 1.0]},
            InterpolatedSpectrum(
                wavelengths=[500.0, 600.0] * ureg.nm,
                values=[0.0, 1.0],
            ),
            None,
        ),
        (
            {
                "quantity": "irradiance",
                "wavelengths": [500.0, 600.0],
                "values": [0.0, 1.0],
            },
            InterpolatedSpectrum(
                quantity=PhysicalQuantity.IRRADIANCE,
                wavelengths=[500.0, 600.0] * ureg.nm,
                values=[0.0, 1.0] * ureg("W/m^2/nm"),
            ),
            "W/m^2/nm",
        ),
        (
            {
                "quantity": "irradiance",
                "wavelengths": [500.0, 600.0],
                "values": [0.0, 1.0] * ureg.kg,
            },
            UnitsError,
            None,
        ),
        (
            {
                "quantity": "irradiance",
                "wavelengths": [500.0, 600.0] * ureg.s,
                "values": [0.0, 1.0],
            },
            UnitsError,
            None,
        ),
        (
            {
                "wavelengths": [500.0, 600.0],
                "values": [0.0, 1.0] * ureg.dimensionless,
            },
            InterpolatedSpectrum(
                quantity=None,
                wavelengths=[500.0, 600.0] * ureg.nm,
                values=[0.0, 1.0] * ureg.dimensionless,
            ),
            None,
        ),
    ],
    ids=[
        "no_args",
        "missing_args_1",
        "missing_args_2",
        "array_shape_mismatch",
        "no_quantity",
        "no_units",
        "inconsistent_units_1",
        "inconsistent_units_2",
        "quantity_values_but_no_quantity",
    ],
)
def test_interpolated_construct(modes_all, tested, expected, units):
    if isinstance(expected, InterpolatedSpectrum):
        s = InterpolatedSpectrum(**tested)

        assert s.quantity is expected.quantity
        assert np.all(s.values == expected.values)

        if isinstance(s.values, pint.Quantity):
            assert isinstance(s.values.magnitude, np.ndarray)
        else:
            assert isinstance(s.values, np.ndarray)

    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            InterpolatedSpectrum(**tested)

    else:
        raise RuntimeError


@pytest.mark.parametrize("quantity", ["dimensionless", None])
@pytest.mark.parametrize(
    "wmin, wmax, expected, isclose_kwargs",
    [
        # Easy case: integrate over full interval
        (500.0 * ureg.nm, 600.0 * ureg.nm, 50.0 * ureg.nm, {"rtol": 1e-10}),
        # Min or max falls in-between two coordinate values
        (540.0 * ureg.nm, 600.0 * ureg.nm, 42.0 * ureg.nm, {"rtol": 1e-10}),
        (550.0 * ureg.nm, 590.0 * ureg.nm, 28.0 * ureg.nm, {"rtol": 1e-10}),
        (540.0 * ureg.nm, 590.0 * ureg.nm, 32.5 * ureg.nm, {"rtol": 1e-10}),
        (530.0 * ureg.nm, 540.0 * ureg.nm, 3.5 * ureg.nm, {"rtol": 1e-10}),
        # Integrating on an interval not intersecting the support yields 0
        (400.0 * ureg.nm, 450.0 * ureg.nm, 0.0 * ureg.nm, {"atol": 1e-10}),
        (400.0 * ureg.nm, 500.0 * ureg.nm, 0.0 * ureg.nm, {"atol": 1e-10}),
        (650.0 * ureg.nm, 700.0 * ureg.nm, 0.0 * ureg.nm, {"atol": 1e-10}),
        (600.0 * ureg.nm, 700.0 * ureg.nm, 0.0 * ureg.nm, {"atol": 1e-10}),
        # Integrating on an interval covering the whole support yields correct
        # integral values
        (450.0 * ureg.nm, 650.0 * ureg.nm, 50.0 * ureg.nm, {"rtol": 1e-10}),
        (500.0 * ureg.nm, 650.0 * ureg.nm, 50.0 * ureg.nm, {"rtol": 1e-10}),
        (450.0 * ureg.nm, 600.0 * ureg.nm, 50.0 * ureg.nm, {"rtol": 1e-10}),
    ],
)
def test_interpolated_integral(
    mode_mono, quantity, wmin, wmax, expected, isclose_kwargs
):
    s = InterpolatedSpectrum(
        wavelengths=[500.0, 525.0, 550.0, 575.0, 600.0],
        values=[0.0, 0.25, 0.5, 0.75, 1.0],
        quantity=quantity,
    )

    integral = s.integral(wmin, wmax)

    np.testing.assert_allclose(
        integral.m_as(ureg.nm), expected.m_as(ureg.nm), **isclose_kwargs
    )


@pytest.mark.parametrize(
    "quantity, values, w, expected",
    [
        (
            "dimensionless",
            [0.0, 1.0],
            [450.0, 500.0, 550.0, 600.0, 650.0] * ureg.nm,
            [0.0, 0.0, 0.5, 1.0, 0.0],
        ),
        (
            "collision_coefficient",
            [0.0, 1.0],
            [450.0, 500.0, 550.0, 600.0, 650.0] * ureg.nm,
            [0.0, 0.0, 0.5, 1.0, 0.0] * ureg.m**-1,
        ),
    ],
)
def test_interpolated_eval_mono(mode_mono, quantity, values, w, expected):
    # No quantity, unitless value
    eval = InterpolatedSpectrum(
        quantity=quantity, values=values, wavelengths=[500.0, 600.0]
    ).eval_mono(w)
    assert np.all(expected == eval)
    assert isinstance(eval, pint.Quantity)


def test_interpolated_eval_mono_wavelengths_decreasing(mode_mono):
    # decreasing wavelength values
    with pytest.raises(ValueError):
        InterpolatedSpectrum(
            quantity="dimensionless",
            values=[0.0, 1.0],
            wavelengths=[600.0, 500.0],
        )


def test_interpolated_eval(modes_all):
    if eradiate.mode().is_mono:
        si = SpectralIndex.new(w=550.0 * ureg.nm)
        expected = 0.5

    elif eradiate.mode().is_ckd:
        si = SpectralIndex.new(w=550.0 * ureg.nm, g=0)
        expected = 0.5

    else:
        raise NotImplementedError

    # Spectrum performs linear interpolation and yields units consistent with
    # quantity
    spectrum = InterpolatedSpectrum(
        quantity="dimensionless",
        wavelengths=[500.0, 600.0],
        values=[0.0, 1.0],
    )
    assert spectrum.eval(si) == expected * spectrum.values.units

    spectrum = InterpolatedSpectrum(
        quantity="irradiance", wavelengths=[500.0, 600.0], values=[0.0, 1.0]
    )
    assert spectrum.eval(si) == expected * spectrum.values.units

    # If no quantity is specified, the evaluation routine returns a
    # unitless value
    spectrum = InterpolatedSpectrum(
        quantity=None,
        wavelengths=[500.0, 600.0],
        values=[0.0, 1.0],
    )
    actual = spectrum.eval(si)
    assert actual == expected
    assert not isinstance(actual, pint.Quantity)


def test_interpolated_kernel_dict(modes_all_mono):
    # Instantiate from factory
    with ucc.override({"radiance": "W/m^2/sr/nm"}):
        spectrum = spectrum_factory.convert(
            {
                "type": "interpolated",
                "quantity": "radiance",
                "wavelengths": [0.5, 0.6],
                "wavelengths_units": "micron",
                "values": [0.0, 1.0],
            }
        )

    with uck.override({"radiance": "kW/m^2/sr/nm"}):
        # Produced kernel dict and params are valid
        mi_wrapper = check_scene_element(spectrum, mi.Texture)
        # Unit scaling is properly applied
        assert np.isclose(mi_wrapper.parameters["value"], 5e-4)


def test_interpolated_from_dataarray(mode_mono):
    da = eradiate.data.load_dataset(
        "spectra/reflectance/lambertian_soil.nc"
    ).reflectance.sel(brightness="darkest")
    spectrum = InterpolatedSpectrum.from_dataarray(quantity="reflectance", dataarray=da)
    assert np.all(spectrum.wavelengths.m_as(da.w.attrs["units"]) == da.w.values)
    assert np.all(spectrum.values.m_as(da.attrs["units"]) == da.values)


@pytest.mark.parametrize(
    "interpolated_kwargs, expected_selected_wcenters",
    [
        (
            {
                "wavelengths": np.linspace(500.0, 600.0, 11),
                "values": ([0] + [1] * 9 + [0]) * ureg.dimensionless,
            },
            np.arange(505, 596, 10) * ureg.nm,
        ),
        (
            {
                "wavelengths": np.linspace(500.0, 600.0, 11),
                "values": ([np.finfo(float).eps] + [1] * 9 + [np.finfo(float).eps])
                * ureg.dimensionless,
            },
            np.arange(495, 596, 10) * ureg.nm,
        ),
        (
            {
                "wavelengths": np.linspace(500.0, 600.0, 11),
                "values": ([0, 0] + [1] * 7 + [0, 0]) * ureg.dimensionless,
            },
            np.arange(515, 586, 10) * ureg.nm,
        ),
        (
            {
                "wavelengths": np.linspace(500.0, 600.0, 11),
                "values": [0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0] * ureg.dimensionless,
            },
            [515, 525, 535, 555, 565, 575, 585] * ureg.nm,
        ),
        (
            {
                "wavelengths": np.linspace(500.0, 600.0, 11),
                "values": [0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0] * ureg.dimensionless,
            },
            [515, 525, 555, 565, 575, 585, 595] * ureg.nm,
        ),
        (
            {
                "wavelengths": np.linspace(505.0, 605.0, 11),
                "values": ([0, 0] + [1] * 7 + [0, 0]) * ureg.dimensionless,
            },
            np.arange(515, 596, 10) * ureg.nm,
        ),
    ],
    ids=[
        "connected_1",
        "connected_2",
        "connected_3",
        "nonconnected_1",
        "nonconnected_2",
        "connected_misaligned",
    ],
)
def test_interpolated_select_ckd(interpolated_kwargs, expected_selected_wcenters):
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(start=280.0, stop=2400.0, step=10.0)
    srf = InterpolatedSpectrum(**interpolated_kwargs)
    selected = srf.select_in(binset)
    assert np.allclose(selected.wcenters, expected_selected_wcenters)


def test_interpolated_select_ckd_inaccurate_bin_bounds():
    # Inaccurate bin bounds (due to numerical precision issues) are corrected by
    # the bin selection method implementation
    binset = BinSet(
        bins=[
            Bin(wmin=500.0, wmax=550.0),
            Bin(wmin=550.0 + np.finfo("float16").eps, wmax=600.0),
            # we use float16 so that even on Windows (where the default float
            # type is float32), this test still works as expected
        ]
    )
    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(500.0, 600.0, 11),
        values=([0, 0] + [1] * 7 + [0, 0]) * ureg.dimensionless,
    )
    selected = srf.select_in(binset).wcenters
    assert np.allclose(selected, [525.0, 575.0] * ureg.nm)



import numpy as np
import pytest

import eradiate
from eradiate.data import data_store
from eradiate.experiments import CanopyExperiment
from eradiate.test_tools.regression import Chi2Test
from eradiate.units import unit_registry as ureg


@pytest.mark.regression
def test_het06_brfpp(mode_mono_double, artefact_dir, session_timestamp):
    r"""
    Coniferous forest (HET06) regression test
    =========================================

    This is a regression test, which compares the simulation results of the
    current branch to an older reference version.

    Rationale
    ---------

    This test case implements a basic canopy scene:

    * Surface with lambertian reflectance
    * No atmosphere
    * Three dimensional canopy

    Parameters

    * Surface: Square surface with labmertian BSDF with :math:`r = 0.86`
    * Canopy: Coniferous trees made up of a conical leaf cloud and a cylindrical
      trunk.

      * Leaf reflectance: 0.08
      * Leaf transmittance: 0.03
      * Trunk reflectance: 0.14

      Disk and tree positioning follow the HET06 scenario of the RAMI-3 benchmark
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points,
      10000 samples per pixel)

    Expected behaviour
    ------------------

    This test uses the Chi-squared criterion with a threshold of 0.05.

    """

    tree_pos_path = data_store.fetch(
        "tests/regression_test_specifications/het06/het06_scene.def"
    )

    exp = CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": tree_pos_path,
                    "canopy_element": {
                        "type": "abstract_tree",
                        "trunk_height": 1.5 * ureg.m,
                        "trunk_radius": 0.15 * ureg.m,
                        "trunk_reflectance": 0.14,
                        "leaf_cloud_extra_offset": [0.0, 0.0, 0.0] * ureg.m,
                        "id": "conifer_tree",
                        "leaf_cloud": {
                            "construct": "cone",
                            "l_vertical": 12.0 * ureg.m,
                            "leaf_radius": 0.05,
                            "radius": 1.8 * ureg.m,
                            "n_leaves": 6480,
                            "leaf_reflectance": 0.08,
                            "leaf_transmittance": 0.03,
                            "id": "conical_leaf_cloud",
                        },
                    },
                }
            ],
            "size": [500, 500, 13.5] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.86},
        padding=0,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 10000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -250 * ureg.m,
                    "xmax": 250 * ureg.m,
                    "ymin": -250 * ureg.m,
                    "ymax": 250 * ureg.m,
                    "z": 6.75 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 40 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1},
    )

    result = eradiate.run(exp)

    test = Chi2Test(
        name=f"{session_timestamp:%Y%m%d-%H%M%S}-het06.nc",
        value=result,
        reference="tests/regression_test_references/het06_brfpp_ref.nc",
        threshold=0.05,
        archive_dir=artefact_dir,
    )

    assert test.run()



import numpy as np
import pytest

import eradiate
from eradiate.experiments import AtmosphereExperiment, CanopyExperiment
from eradiate.scenes.bsdfs import LambertianBSDF
from eradiate.scenes.illumination import DirectionalIllumination
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.scenes.spectra import SolarIrradianceSpectrum
from eradiate.units import unit_registry as ureg


@pytest.mark.parametrize("cls", ["onedim", "rami"])
@pytest.mark.parametrize(
    "wavelengths",
    [
        [500.0] * ureg.nm,
        [500.0 + 100 * i for i in range(11)] * ureg.nm,
        [500.0 + 100.0 * i for i in range(10, -1, -1)] * ureg.nm,
    ],
    ids=["single", "multiple_ascending", "multiple_descending"],
)
@pytest.mark.parametrize("irradiance", ["uniform", "solar"])
def test_spectral_loop(mode_mono, cls, wavelengths, irradiance):
    """
    Spectral loop
    =============

    This test case checks if the spectral loop implementation and
    post-processing behaves as intended.

    Rationale
    ---------

    The scene consists of a square Lambertian surface with reflectance
    :math:`\\rho = 1` illuminated by a directional emitter positioned at the
    zenith. The reflected radiance is computed in monochromatic mode for a
    single wavelength and for multiple wavelength. The experiment is repeated:

    * for all solver applications supporting this type of scene;
    * for several spectral configurations (single wavenlength, multiple
      ascending, multiple descending);
    * for several irradiance spectra (uniform, solar).

    Expected behaviour
    ------------------

    All BRF values are equal to 1.
    """
    if cls == "onedim":
        cls_exp = AtmosphereExperiment
        kwargs = {"atmosphere": None}
    elif cls == "rami":
        cls_exp = CanopyExperiment
        kwargs = {"canopy": None}
    else:
        raise ValueError

    if irradiance == "uniform":
        illumination = DirectionalIllumination(irradiance=1.0)
    elif irradiance == "solar":
        illumination = DirectionalIllumination(irradiance=SolarIrradianceSpectrum())

    exp = cls_exp(
        surface=LambertianBSDF(reflectance=1.0),
        illumination=illumination,
        measures=[
            MultiDistantMeasure(
                spp=1,
                srf={"type": "multi_delta", "wavelengths": wavelengths},
            )
        ],
        **kwargs,
    )

    results = eradiate.run(exp)
    assert np.allclose(np.squeeze(results.brf.values), 1.0)



# This script walks through plugin files and extracts documentation that should
# go into the reference manual.
import glob
import re
import typing as t
from pathlib import Path
from textwrap import dedent

# Section titles (order matters)
SECTIONS = {
    "shapes": "Shapes",
    "bsdfs": "BSDFs",
    "media": "Media",
    "phase": "Phase functions",
    "emitters": "Emitters",
    "sensors": "Sensors",
    "textures": "Textures",
    "spectra": "Spectra",
    "volumes": "Volume data sources",
    "integrators": "Integrators",
    "samplers": "Samplers",
    "films": "Films",
    "rfilters": "Reconstruction filters",
}

# Plugin list
PLUGIN_ORDERS = {
    "bsdfs": [
        "bilambertian",
        "rpv",
    ],
    "sensors": [
        "distantflux",
        "hdistant",
        "mdistant",
        "mradiancemeter",
    ],
    "emitters": [
        "astroobject",
    ],
}


def write_if_modified(filename, content):
    filename.parent.mkdir(parents=True, exist_ok=True)

    try:
        with open(filename, "r") as f:
            existing = f.read()
    except OSError:
        existing = None

    if existing == content:
        print(f"Skipping unchanged '{filename.name}'")

    else:
        print(f"Generating '{filename.name}'")
        with open(filename, "w") as f:
            f.write(content)


def underline(s: str, c: str = "-"):
    return c * len(s)


def extract_plugin_docs(filename: Path) -> t.List[str]:
    result = []

    with open(filename) as in_file:
        inheader = False

        for line in in_file.readlines():
            match = re.match(r"^/\*\*! ?(.*)$", line)
            if match is not None:
                line = match.group(1).replace("%", "\%")
                result.append(line.rstrip())
                inheader = True
                continue

            if not inheader:
                continue

            if re.search(r"^[\s\*]*\*/$", line):
                inheader = False
                continue

            result.append(line.rstrip())

    return ("\n".join(result)).lstrip()


def make_toc(section, title) -> str:
    return dedent(
        rf"""
        .. _sec-reference_plugins-{section}:

        {title}
        {underline(title)}

        .. toctree::
           :maxdepth: 1
           :glob:

           ../plugins/{section}/*
        """
    )


def generate():
    root_dir = Path(__file__).absolute().parent.parent
    plugin_dir = root_dir / "ext/mitsuba/src/eradiate_plugins"
    out_dir = root_dir / "docs/rst/reference_plugins/generated"
    print(f"Generating plugin docs in '{out_dir}'")

    for section, section_title in SECTIONS.items():
        try:
            plugin_order = PLUGIN_ORDERS[section]
        except KeyError:
            plugin_order = []

        # Get plugin filenames
        section_dir = plugin_dir / section
        plugin_list = [
            Path(filename).stem for filename in glob.glob(f"{section_dir}/*.cpp")
        ]

        if not plugin_list:
            print(f"No plugin in section '{section}'")
            continue

        # Create TOC page
        section_rst = out_dir / f"toctrees/{section}.rst"
        write_if_modified(section_rst, make_toc(section, section_title))

        # Sort plugin list according to the ordering dict
        def index(x):
            try:
                return plugin_order.index(x)
            except ValueError:
                return 1000

        plugin_list.sort(key=index)

        # Extract plugin docs
        print(f"Extracting plugin docs from '{plugin_dir}' to '{out_dir}'")
        for plugin_name in plugin_list:
            plugin_cpp = plugin_dir / section / f"{plugin_name}.cpp"
            plugin_rst = out_dir / "plugins" / section / f"{plugin_name}.rst"

            print(f"Processing '{Path(plugin_cpp).relative_to(root_dir)}'")
            write_if_modified(plugin_rst, extract_plugin_docs(plugin_cpp))


if __name__ == "__main__":
    generate()



import numpy as np
import pytest
import xarray as xr

from eradiate import KernelContext, data
from eradiate import unit_registry as ureg
from eradiate.radprops import ZGrid
from eradiate.scenes.atmosphere import ParticleLayer, UniformParticleDistribution
from eradiate.scenes.core import traverse
from eradiate.spectral.index import SpectralIndex
from eradiate.test_tools.types import check_scene_element
from eradiate.units import to_quantity

# ------------------------------------------------------------------------------
#                              Fixtures
# ------------------------------------------------------------------------------


@pytest.fixture
def test_dataset_path():
    """Test dataset path fixture."""
    return "tests/radprops/rtmom_aeronet_desert.nc"


# ------------------------------------------------------------------------------
#                                   Tests
# ------------------------------------------------------------------------------


@pytest.mark.parametrize("geometry", ["plane_parallel", "spherical_shell"])
def test_particle_layer_kernel_dict(mode_mono, geometry):
    """
    Kernel dictionary checks
    """
    layer = ParticleLayer(geometry=geometry)
    check_scene_element(layer)


@pytest.mark.parametrize("w", [280.0, 550.0, 1600.0, 2400.0] * ureg.nm)
def test_particle_layer_eval_mono_absorbing_only(
    mode_mono, particle_dataset_absorbing_only, w
):
    """eval methods return expected values for an absorbing-only layer."""
    bottom = 0.0 * ureg.km
    top = 1.0 * ureg.km
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        dataset=particle_dataset_absorbing_only,
        bottom=bottom,
        top=top,
    )
    si = SpectralIndex.new(w=w)
    assert np.allclose(layer.eval_sigma_s(si), 0.0 / ureg.km)
    assert np.allclose(layer.eval_sigma_a(si), 0.2 / ureg.km)
    assert np.allclose(layer.eval_albedo(si).m, 0.0)

    ctx = KernelContext(si=si)
    assert layer.eval_mfp(ctx).magnitude == np.inf


@pytest.mark.parametrize("w", [280.0, 550.0, 1600.0, 2400.0] * ureg.nm)
def test_particle_layer_eval_mono_scattering_only(
    mode_mono, particle_dataset_scattering_only, w
):
    """eval methods return expected values for a scattering-only layer."""
    bottom = 0.0 * ureg.km
    top = 1.0 * ureg.km
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        dataset=particle_dataset_scattering_only,
        bottom=bottom,
        top=top,
    )
    si = SpectralIndex.new(w=w)
    assert np.allclose(layer.eval_sigma_s(si), 0.2 / ureg.km)
    assert np.allclose(layer.eval_sigma_a(si), 0.0 / ureg.km)
    assert np.allclose(layer.eval_albedo(si).m, 1.0)

    ctx = KernelContext(si=si)
    assert np.isclose(layer.eval_mfp(ctx), 5.0 * ureg.km)


@pytest.mark.parametrize("w", [280.0, 550.0, 1600.0, 2400.0] * ureg.nm)
def test_particle_layer_eval_mono(mode_mono, particle_dataset_test, w):
    """
    eval_* methods return expected values for a scattering and absorbing layer.
    """
    bottom = 0.0 * ureg.km
    top = 1.0 * ureg.km
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        dataset=particle_dataset_test,
        tau_ref=1.0,
        bottom=bottom,
        top=top,
    )
    si = SpectralIndex.new(w=w)
    assert np.allclose(layer.eval_sigma_t(si), 1.0 / ureg.km)
    assert np.allclose(layer.eval_sigma_s(si), 0.8 / ureg.km)
    assert np.allclose(layer.eval_sigma_a(si), 0.2 / ureg.km)
    assert np.allclose(layer.eval_albedo(si).m, 0.8)

    ctx = KernelContext(si=si)
    assert np.isclose(layer.eval_mfp(ctx), 1.25 * ureg.km)


@pytest.mark.parametrize("w", [280.0, 550.0, 1600.0, 2400.0] * ureg.nm)
def test_particle_layer_eval_ckd_absorbing_only(
    mode_ckd, particle_dataset_absorbing_only, w
):
    """eval methods return expected values for an absorbing-only layer."""
    bottom = 0.0 * ureg.km
    top = 1.0 * ureg.km
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        dataset=particle_dataset_absorbing_only,
        bottom=bottom,
        top=top,
    )
    si = SpectralIndex.new(w=w)
    assert np.allclose(layer.eval_sigma_s(si), 0.0 / ureg.km)
    assert np.allclose(layer.eval_sigma_a(si), 0.2 / ureg.km)
    assert np.allclose(layer.eval_albedo(si).m, 0.0)

    ctx = KernelContext(si=si)
    assert layer.eval_mfp(ctx).magnitude > 0.0


@pytest.mark.parametrize("w", [280.0, 550.0, 1600.0, 2400.0] * ureg.nm)
def test_particle_layer_eval_ckd_scattering_only(
    mode_ckd, particle_dataset_scattering_only, w
):
    """eval methods return expected values for a scattering-only layer."""
    bottom = 0.0 * ureg.km
    top = 1.0 * ureg.km
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        dataset=particle_dataset_scattering_only,
        bottom=bottom,
        top=top,
    )
    si = SpectralIndex.new(w=w)
    assert np.allclose(layer.eval_sigma_s(si), 0.2 / ureg.km)
    assert np.allclose(layer.eval_sigma_a(si).m, 0.0)
    assert np.allclose(layer.eval_albedo(si).m, 1.0)

    ctx = KernelContext(si=si)
    assert layer.eval_mfp(ctx).magnitude > 0.0


@pytest.mark.parametrize("w", [280.0, 550.0, 1600.0, 2400.0] * ureg.nm)
def test_particle_layer_eval_ckd(mode_ckd, particle_dataset_test, w):
    """eval methods return expected values for a scattering-only layer."""
    bottom = 0.0 * ureg.km
    top = 1.0 * ureg.km
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        dataset=particle_dataset_test,
        tau_ref=1.0,
        bottom=bottom,
        top=top,
    )
    si = SpectralIndex.new(w=w)
    assert np.allclose(layer.eval_sigma_t(si), 1.0 / ureg.km)
    assert np.allclose(layer.eval_sigma_s(si), 0.8 / ureg.km)
    assert np.allclose(layer.eval_sigma_a(si), 0.2 / ureg.km)
    assert np.allclose(layer.eval_albedo(si).m, 0.8)

    ctx = KernelContext(si=si)
    assert layer.eval_mfp(ctx) == 1.25 * ureg.km


def test_particle_layer_construct_basic():
    """Construction succeeds with basic parameters."""
    assert ParticleLayer()


def test_particle_layer_scale(modes_all_single):
    """Scale parameter propagates to kernel dict and latter can be loaded."""
    particle_layer = ParticleLayer(id="atmosphere", scale=2.0)
    template, _ = traverse(particle_layer)
    assert template["medium_atmosphere.scale"] == 2.0


def test_particle_layer_construct_attrs(test_dataset_path):
    """Assigns parameters to expected values."""
    bottom = ureg.Quantity(1.2, "km")
    top = ureg.Quantity(1.8, "km")
    tau_ref = ureg.Quantity(0.3, "dimensionless")
    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "ground_altitude": bottom,
            "toa_altitude": top,
        },
        bottom=bottom,
        top=top,
        distribution=UniformParticleDistribution(),
        tau_ref=tau_ref,
        dataset=test_dataset_path,
    )
    assert layer.bottom_altitude == bottom
    assert layer.top_altitude == top
    assert isinstance(layer.distribution, UniformParticleDistribution)
    assert layer.tau_ref == tau_ref
    assert isinstance(layer.dataset, xr.Dataset)


def test_particle_layer_altitude_units():
    """Accept different units for bottom and top altitudes."""
    assert ParticleLayer(bottom=1.0 * ureg.km, top=2000.0 * ureg.m)


def test_particle_layer_invalid_bottom_top():
    """Raises when 'bottom' is larger that 'top'."""
    with pytest.raises(ValueError):
        ParticleLayer(top=1.2 * ureg.km, bottom=1.8 * ureg.km)


def test_particle_layer_invalid_tau_ref():
    """Raises when 'tau_ref' is invalid."""
    with pytest.raises(ValueError):
        ParticleLayer(
            bottom=1.2 * ureg.km,
            top=1.8 * ureg.km,
            tau_ref=-0.1 * ureg.dimensionless,
        )


def test_particle_layer_eval_radprops_format(
    modes_all_single,
    test_dataset_path,
):
    """
    Method 'eval_radprops' returns dataset with expected datavars and coords.
    """
    layer = ParticleLayer(dataset=test_dataset_path)
    si = SpectralIndex.new()
    ds = layer.eval_radprops(si)
    expected_data_vars = ["sigma_t", "albedo"]
    expected_coords = ["z_layer"]
    assert all([coord in ds.coords for coord in expected_coords]) and all(
        [var in ds.data_vars for var in expected_data_vars]
    )


@pytest.mark.parametrize(
    "tau_ref",
    np.array([0.6, 1.0, 2.5]) * ureg.dimensionless,
)
def test_particle_layer_eval_radprops(mode_mono, test_dataset_path, tau_ref):
    layer = ParticleLayer(
        dataset=test_dataset_path,
        bottom=0.5 * ureg.km,  # arbitrary
        top=3.0 * ureg.km,  # arbitrary
        distribution={"type": "uniform"},
        tau_ref=tau_ref,
    )

    # compute optical thickness at reference wavelength from layer's radprops
    # and check it matches the input tau_ref
    si = SpectralIndex.new(w=layer.w_ref)
    radprops = layer.eval_radprops(si)
    delta_z = layer.geometry.zgrid.layer_height

    with xr.set_options(keep_attrs=True):
        tau = to_quantity(radprops.sigma_t.sum()) * delta_z

    assert np.isclose(tau, tau_ref)


@pytest.mark.parametrize("distribution", ["uniform", "gaussian", "exponential"])
@pytest.mark.parametrize(
    "tau_ref",
    np.array([0.1, 0.5, 1.0, 2.0, 5.0]) * ureg.dimensionless,
)
def test_particle_layer_eval_sigma_t_impl(
    mode_mono, tau_ref, distribution, test_dataset_path
):
    r"""
    Spectral dependency of extinction is accounted for.

    If :math:`\sigma_t(\lambda)` denotes the extinction coefficient at the
    wavelength :math:`\lambda`, then the optical thickness of a uniform
    particle layer is :math:`\tau(\lambda) = \sigma_t(\lambda) \, \Delta z`
    where :math:`\Delta z` is the layer's thickness.
    It follows that:

    .. math::

       \frac{\tau(\lambda)}{\tau(\lambda_{\mathrm{ref}})} =
       \frac{\sigma(\lambda)}{\sigma(\lambda_{\mathrm{ref}})}

    which is what we assert in this test.
    """
    w_ref = 550 * ureg.nm
    bottom = 1.0 * ureg.km  # arbitrary
    top = 4.0 * ureg.km  # arbitrary
    n_wavelengths = 3
    n_layers = 10
    wavelengths = np.linspace(500.0, 1500.0, n_wavelengths) * ureg.nm
    zgrid = ZGrid(np.linspace(0, 5, n_layers + 1) * ureg.km)

    layer = ParticleLayer(
        geometry={
            "type": "plane_parallel",
            "toa_altitude": zgrid.levels[-1],
            "zgrid": zgrid,
        },
        dataset=test_dataset_path,
        bottom=bottom,
        top=top,
        distribution=distribution,
        w_ref=w_ref,
        tau_ref=tau_ref,
    )

    # layer optical thickness @ current wavelength
    sigma_t = layer._eval_sigma_t_impl(wavelengths, layer.geometry.zgrid)
    assert sigma_t.units.is_compatible_with(ureg("m**-1"))
    assert sigma_t.shape == (n_wavelengths, n_layers)
    tau = np.sum(
        sigma_t * layer.geometry.zgrid.layer_height,
        axis=1,
    )  # Integrate sigma_t * dz vs space coordinate using rectangle method

    # data set extinction @ running and reference wavelength
    ds = data.load_dataset(test_dataset_path)
    w_units = ureg(ds.w.attrs["units"])
    sigma_t = to_quantity(ds.sigma_t.interp(w=wavelengths.m_as(w_units)))
    sigma_t_ref = to_quantity(ds.sigma_t.interp(w=w_ref.m_as(w_units)))

    # the spectral dependence of the optical thickness and extinction
    # coefficient match, so the below ratios must match
    result = (tau / tau_ref).m_as(ureg.dimensionless)
    expected = (sigma_t / sigma_t_ref).m_as(ureg.dimensionless)
    np.testing.assert_allclose(result, expected)


@pytest.mark.parametrize(
    "has_absorption, has_scattering, expected",
    [
        (
            True,
            True,
            {
                "albedo": 0.97331806,
                "sigma_t": 1.0,
                "sigma_a": 0.02668194,
                "sigma_s": 0.97331806,
            },
        ),
        (
            False,
            True,
            {
                "albedo": 1.0,
                "sigma_t": 0.97331806,
                "sigma_a": 0.0,
                "sigma_s": 0.97331806,
            },
        ),
        (
            True,
            False,
            {
                "albedo": 0.0,
                "sigma_t": 0.02668194,
                "sigma_a": 0.02668194,
                "sigma_s": 0.0,
            },
        ),
        (False, False, "raise"),
    ],
)
def test_particle_layer_switches(mode_mono, has_absorption, has_scattering, expected):
    try:
        bottom = 0.0 * ureg.km
        top = 1.0 * ureg.km
        particle_layer = ParticleLayer(
            geometry={
                "type": "plane_parallel",
                "ground_altitude": bottom,
                "toa_altitude": top,
            },
            tau_ref=1.0,
            has_absorption=has_absorption,
            has_scattering=has_scattering,
        )
        zgrid = particle_layer.geometry.zgrid
        w = 550.0 * ureg.nm

        np.testing.assert_allclose(
            particle_layer.eval_albedo_mono(w, zgrid).m_as(ureg.dimensionless),
            expected["albedo"],
        )
        np.testing.assert_allclose(
            particle_layer.eval_sigma_t_mono(w, zgrid).m_as("km^-1"),
            expected["sigma_t"],
        )
        np.testing.assert_allclose(
            particle_layer.eval_sigma_a_mono(w, zgrid).m_as("km^-1"),
            expected["sigma_a"],
        )
        np.testing.assert_allclose(
            particle_layer.eval_sigma_s_mono(w, zgrid).m_as("km^-1"),
            expected["sigma_s"],
        )

    except ValueError:
        assert expected == "raise"



from __future__ import annotations

import enum
import importlib.resources
from pathlib import Path

from dynaconf import Dynaconf, Validator

from ._env import ENV, SOURCE_DIR
from ..frame import AzimuthConvention


class ProgressLevel(enum.IntEnum):
    """
    An enumeration defining valid progress levels.

    This is an integer enumeration, meaning that levels can be compared to
    numerics.
    """

    NONE = 0  #: No progress
    SPECTRAL_LOOP = enum.auto()  #: Up to spectral loop level progress
    KERNEL = enum.auto()  #: Up to kernel level progress


def _default_download_dir():
    return (
        Path("./eradiate_downloads").absolute().resolve()
        if SOURCE_DIR is None
        else SOURCE_DIR / ".eradiate_downloads"
    )


DEFAULTS = importlib.resources.files("eradiate.config").joinpath(f"{ENV}.toml")

#: Main settings data structure. See the `Dynaconf documentation <https://www.dynaconf.com/>`_
#: for details.
settings = Dynaconf(
    settings_files=[DEFAULTS, "eradiate.toml"],
    envvar_prefix="ERADIATE",
    merge_enabled=True,
    validate_on_update=True,
    validators=[
        Validator(
            "AZIMUTH_CONVENTION",
            cast=lambda x: (AzimuthConvention[x.upper()] if isinstance(x, str) else x),
        ),
        Validator("DATA_STORE_URL", cast=str),
        Validator(
            "DOWNLOAD_DIR",
            default=_default_download_dir(),
            cast=lambda x: Path(x).resolve(),
        ),
        Validator("OFFLINE", cast=bool),
        Validator(
            "PROGRESS",
            cast=lambda x: (
                ProgressLevel[x.upper()] if isinstance(x, str) else ProgressLevel(x)
            ),
        ),
        Validator("SMALL_FILES_REGISTRY_URL", cast=str),
        Validator("SMALL_FILES_REGISTRY_REVISION", cast=str),
    ],
)



from importlib import import_module
from pathlib import Path
from textwrap import dedent, indent

# Auto-generation disclaimer text
HEADER = dedent(
    """
    ..
      This file was automatically generated by docs/generate_rst_api.py. The

          make docs-rst-api

      target automates this process.
    """
).strip()


def write_if_modified(filename, content):
    filename.parent.mkdir(parents=True, exist_ok=True)

    try:
        with open(filename, "r") as f:
            existing = f.read()
    except OSError:
        existing = None

    if existing == content:
        print(f"Skipping unchanged '{filename.name}'")

    else:
        print(f"Generating '{filename.name}'")
        with open(filename, "w") as f:
            f.write(content)


# List of (module, variable) pairs
FACTORIES = [
    ("eradiate.radprops", "rad_profile_factory"),
    ("eradiate.scenes.atmosphere", "atmosphere_factory"),
    ("eradiate.scenes.atmosphere", "particle_distribution_factory"),
    ("eradiate.scenes.biosphere", "biosphere_factory"),
    ("eradiate.scenes.bsdfs", "bsdf_factory"),
    ("eradiate.scenes.illumination", "illumination_factory"),
    ("eradiate.scenes.integrators", "integrator_factory"),
    ("eradiate.scenes.measure", "measure_factory"),
    ("eradiate.scenes.phase", "phase_function_factory"),
    ("eradiate.scenes.shapes", "shape_factory"),
    ("eradiate.scenes.spectra", "spectrum_factory"),
    ("eradiate.scenes.surface", "surface_factory"),
]


def factory_data_docs(modname, varname, uline="="):
    """
    Return rst code for a factory instance located at modname.varname.
    """
    factory = getattr(import_module(modname), varname)
    fullname = f"{modname}.{varname}"
    underline = uline * len(fullname)

    table_header = ".. list-table::\n   :widths: 25 75"
    table_rows = "\n".join(
        [
            f"   * - ``{key}``\n     - :class:`{factory.get_type(key).__name__}`"
            for key in sorted(factory.registry.keys())
        ]
    )
    table = "\n".join([table_header, "", table_rows])

    return f"""{fullname}
{underline}

.. data:: {modname}.{varname}
   :annotation:

   Instance of :class:`{factory.__class__.__module__}.{factory.__class__.__qualname__}`

   .. rubric:: Registered types

{indent(table, "   ")}

""".lstrip()


def generate_factory_docs():
    """
    Generate rst documents to display factory documentation.
    """
    outdir = Path(__file__).parent.absolute() / "rst/reference_api/generated/factory"
    print(f"Generating factory docs in '{outdir}'")

    for modname, varname in FACTORIES:
        generated = factory_data_docs(modname, varname)
        content = "\n".join([HEADER, "", generated, ""])
        write_if_modified(outdir / f"{modname}.{varname}.rst", content)


if __name__ == "__main__":
    generate_factory_docs()



"""Test cases of the _molecular module."""

import mitsuba as mi
import numpy as np
import numpy.testing as npt
import pytest

from eradiate import unit_registry as ureg
from eradiate.contexts import KernelContext
from eradiate.scenes.atmosphere import MolecularAtmosphere
from eradiate.scenes.core import Scene, traverse
from eradiate.spectral import CKDSpectralIndex


def test_molecular_atmosphere_default_mono(mode_mono):
    atmosphere = MolecularAtmosphere()
    si = atmosphere.spectral_set().spectral_indices().__next__()
    template, _ = traverse(atmosphere)
    assert template.render(KernelContext(si=si))


def test_molecular_atmosphere_default_ckd(mode_ckd):
    atmosphere = MolecularAtmosphere()
    si = next(atmosphere.spectral_set().spectral_indices())
    template, _ = traverse(atmosphere)
    assert template.render(KernelContext(si=si))


def test_molecular_atmosphere_scale(
    mode_mono, absorption_database_error_handler_config
):
    atmosphere = MolecularAtmosphere(
        thermoprops={
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0, 120, 121) * ureg.km,
            "additional_molecules": False,
        },
        absorption_data="komodo",
        error_handler_config=absorption_database_error_handler_config,
        scale=2.0,
    )
    template, _ = traverse(atmosphere)
    kernel_dict = template.render(KernelContext())
    assert kernel_dict["medium_atmosphere"]["scale"] == 2.0


def test_molecular_atmosphere_kernel_dict(
    mode_ckd, absorption_database_error_handler_config
):
    """Constructor produces a valid kernel dictionary."""

    atmosphere = MolecularAtmosphere(
        thermoprops={
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0, 80, 41) * ureg.km,
            "additional_molecules": False,
        },
        absorption_data="monotropa",
        error_handler_config=absorption_database_error_handler_config,
        geometry={
            "type": "spherical_shell",
            "ground_altitude": 0 * ureg.km,
            "toa_altitude": 80 * ureg.km,
        },
    )

    sis = [CKDSpectralIndex(w=550.0, g=g) for g in [0.25, 0.75]]
    kernel_context = KernelContext(si=sis[0])

    template, params = traverse(Scene(objects={"atmosphere": atmosphere}))
    mi_scene: mi.Scene = mi.load_dict(template.render(kernel_context))

    # Mono evaluation generates valid parameter update tables
    mi_params: mi.SceneParameters = mi.traverse(mi_scene)

    # for w in eval_w:
    for si in sis:
        ctx = KernelContext(si=si)
        mi_params.update(params.render(ctx))


def test_molecular_atmosphere_switches(
    mode_mono, absorption_database_error_handler_config
):
    # Absorption can be deactivated
    atmosphere = MolecularAtmosphere(
        absorption_data="komodo",
        has_absorption=False,
        error_handler_config=absorption_database_error_handler_config,
    )
    ctx = KernelContext()
    radprops = atmosphere.eval_radprops(ctx.si, optional_fields=True)
    npt.assert_allclose(radprops.sigma_a, 0.0)

    # Scattering can be deactivated
    atmosphere = MolecularAtmosphere(
        absorption_data="komodo",
        thermoprops={
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0.0, 120.0, 121) * ureg.km,
            "additional_molecules": False,
        },
        has_scattering=False,
        error_handler_config=absorption_database_error_handler_config,
    )

    si = next(atmosphere.spectral_set().spectral_indices())
    radprops = atmosphere.eval_radprops(si, optional_fields=True)
    npt.assert_allclose(radprops.sigma_s, 0.0)

    # At least one must be active
    with pytest.raises(ValueError):
        MolecularAtmosphere(
            absorption_data="komodo",
            has_absorption=False,
            has_scattering=False,
            error_handler_config=absorption_database_error_handler_config,
        )



import drjit as dr
import mitsuba as mi
import numpy as np
import pytest

from eradiate import KernelContext
from eradiate import unit_context_config as ucc
from eradiate import unit_context_kernel as uck
from eradiate import unit_registry as ureg
from eradiate.scenes.core import traverse
from eradiate.scenes.shapes import CuboidShape
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs, expected_reflectance",
    [({}, None), ({"bsdf": {"type": "lambertian", "reflectance": 0.0}}, 0.0)],
    ids=["noargs", "args"],
)
def test_cuboid_construct_kernel_dict(modes_all, kwargs, expected_reflectance):
    cuboid = CuboidShape(**kwargs)

    mi_wrapper = check_scene_element(cuboid, mi.Shape)

    if expected_reflectance is not None:
        assert mi_wrapper.parameters["bsdf.reflectance.value"] == expected_reflectance


def test_cuboid_construct_trafo(mode_mono):
    assert CuboidShape(
        edges=[1, 1, 1], to_world=mi.Transform4f.rotate(axis=[1, 1, 1], angle=45)
    )


@pytest.mark.parametrize(
    "kwargs, expected_transform",
    [
        (
            {"edges": [2, 4, 8]},
            [(-1, -2, -4), (1, 2, 4)],
        ),
        (
            {"edges": [2, 2, 2], "center": [1, 1, 1]},
            [(0, 0, 0), (2, 2, 2)],
        ),
        ({"to_world": True}, [(1, 1, 1), (3, 3, 3)]),
        (
            {"edges": [2, 2, 2], "center": [1, 1, 1], "to_world": True},
            [(1, 1, 1), (3, 3, 3)],
        ),
    ],
    ids=["edges", "center", "to_world", "full"],
)
def test_cuboid_params(mode_mono_double, kwargs, expected_transform):
    if "to_world" in kwargs:
        trafo = kwargs.pop("to_world")
        if trafo:
            to_world = mi.ScalarTransform4f.translate((2, 2, 2))
            cuboid = CuboidShape(**kwargs, to_world=to_world)
    else:
        cuboid = CuboidShape(**kwargs)

    template, _ = traverse(cuboid)
    kernel_dict = template.render(ctx=KernelContext())
    to_world = kernel_dict["to_world"]
    assert dr.allclose(
        to_world.transform_affine(mi.Point3f(-1, -1, -1)), expected_transform[0]
    )
    assert dr.allclose(
        to_world.transform_affine(mi.Point3f(1, 1, 1)), expected_transform[1]
    )


def test_cuboid_atmosphere(mode_mono_double):
    with ucc.override(length="km"):
        cuboid = CuboidShape.atmosphere(
            top=1.0,
            bottom=0.0,
            bottom_offset=-0.5,
            width=1.0,
        )
        assert np.allclose(cuboid.edges, [1, 1, 1.5] * ureg.km)
        assert np.allclose(cuboid.center, [0, 0, 0.25] * ureg.km)

    with uck.override(length="m"):
        template, _ = traverse(cuboid)
        kernel_dict = template.render(ctx=KernelContext())
        bbox = mi.load_dict(kernel_dict).bbox()
        assert dr.allclose(bbox.min, [-500, -500, -500])
        assert dr.allclose(bbox.max, [500, 500, 1000])


def test_cuboid_contains(mode_mono):
    cuboid = CuboidShape(center=[0.5, 0.5, 0.5], edges=[1, 1, 1])

    # Works with a single point
    assert cuboid.contains([0.5, 0.5, 0.5])
    assert not cuboid.contains([0.5, 0.5, -0.5])
    assert not cuboid.contains([0.5, 0.5, 0.5] * ureg.km)

    # Works with multiple points
    assert np.all(cuboid.contains([[0.5, 0.5, 0.5], [0.5, -0.5, 0.5]]) == [True, False])

    # to_world
    import mitsuba as mi

    cuboid = CuboidShape(to_world=mi.Transform4f.rotate(angle=45, axis=(0, 0, 1)))

    # Works with a single point
    assert cuboid.contains([0.5, 0.5, 0.5])
    assert not cuboid.contains([1.0, 1.0, 1.0])
    assert not cuboid.contains([0.5, 0.5, 0.5] * ureg.km)

    # Works with multiple points
    assert np.all(cuboid.contains([[0.5, 0.5, 0.5], [1.0, 1.0, 1.0]]) == [True, False])


def test_cuboid_bbox(mode_mono):
    cuboid = CuboidShape(center=[0.5, 0.5, 0.5], edges=[1, 1, 1])
    bbox = cuboid.bbox
    np.testing.assert_array_equal(bbox.min.m_as(ureg.m), [0, 0, 0])
    np.testing.assert_array_equal(bbox.max.m_as(ureg.m), [1, 1, 1])

    import mitsuba as mi

    cuboid = CuboidShape(to_world=mi.ScalarTransform4f.translate((1, 2, 3)))
    bbox = cuboid.bbox
    np.testing.assert_array_equal(bbox.min.m_as(ureg.m), [0, 1, 2])
    np.testing.assert_array_equal(bbox.max.m_as(ureg.m), [2, 3, 4])



import re

import pytest

from eradiate.kernel import (
    InitParameter,
    KernelDictTemplate,
    UpdateMapTemplate,
    UpdateParameter,
)


def test_kernel_dict_template_render():
    template = KernelDictTemplate(
        {
            "foo.bar": 0,
            "bar": InitParameter(lambda ctx: ctx),
            "baz": InitParameter(lambda ctx: ctx),
        }
    )

    assert template.render(ctx=1, nested=True) == {
        "foo": {"bar": 0},
        "bar": 1,
        "baz": 1,
    }

    assert template.render(ctx=1, nested=False) == {"foo.bar": 0, "bar": 1, "baz": 1}


def test_update_map_template_render():
    pmap = UpdateMapTemplate(
        {
            "foo": 0,
            "bar": UpdateParameter(lambda ctx: ctx, UpdateParameter.Flags.GEOMETRIC),
            "baz": UpdateParameter(lambda ctx: ctx, UpdateParameter.Flags.SPECTRAL),
        }
    )

    # If no flags are passed, all params are rendered
    result = pmap.render(ctx=1, flags=UpdateParameter.Flags.ALL)
    assert result["bar"] == 1 and result["baz"] == 1

    # If a flag is passed, only the corresponding params are rendered
    with pytest.raises(ValueError, match=re.escape("Unevaluated parameters: ['bar']")):
        pmap.render(ctx=1, flags=UpdateParameter.Flags.SPECTRAL, drop=False)

    # If drop is set to True, unused parameters are dropped
    result = pmap.render(ctx=1, flags=UpdateParameter.Flags.SPECTRAL, drop=True)
    assert result["baz"] == 1
    assert "bar" not in result


def test_update_map_template_keep_remove():
    template = UpdateMapTemplate(
        {
            "foo": 0,
            "foo.bar": 1,
            "bar": UpdateParameter(lambda ctx: ctx, UpdateParameter.Flags.GEOMETRIC),
            "baz": UpdateParameter(lambda ctx: ctx, UpdateParameter.Flags.SPECTRAL),
        }
    )

    # We can remove or keep selected parameters
    pmap = template.copy()
    pmap.remove(r"foo.*")
    assert pmap.keys() == {"bar", "baz"}

    pmap = template.copy()
    pmap.keep(r"foo.*")
    assert pmap.keys() == {"foo", "foo.bar"}



import numpy as np
import pytest

import eradiate
from eradiate.data import data_store
from eradiate.experiments import CanopyExperiment
from eradiate.test_tools.regression import Chi2Test
from eradiate.units import unit_registry as ureg


@pytest.mark.regression
def test_het01_brfpp(mode_mono_double, artefact_dir, session_timestamp):
    r"""
    Floating spheres (HET01) regression test
    ========================================

    This is a regression test, which compares the simulation results of the
    current branch to an older reference version.

    Rationale
    ---------

    This test case implements a basic canopy scene:

    * Surface with lambertian reflectance
    * No atmosphere
    * Three dimensional canopy

    Parameters

    * Surface: Square surface with labmertian BSDF with :math:`r = 0.159`
    * Canopy: Floating spheres made up of disks with bilambertian bsdf model
      Leaf reflectance is 0.4957, transmittance is 0.4409.
      Disk and sphere positioning follow the HET01 scenario of the RAMI-3 benchmark
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points,
      10000 samples per pixel)

    Expected behaviour
    ------------------

    This test uses the Chi-squared criterion with a threshold of 0.05.

    """

    leaf_spec_path = data_store.fetch(
        "tests/regression_test_specifications/het01/het01_UNI_sphere.def"
    )
    leaf_pos_path = data_store.fetch(
        "tests/regression_test_specifications/het01/het01_UNI_instances.def"
    )

    exp = CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path,
                        "leaf_reflectance": 0.4957,
                        "leaf_transmittance": 0.4409,
                        "id": "spherical_leaf_cloud",
                    },
                },
            ],
            "size": [100, 100, 30] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.159},
        padding=20,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 10000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -50 * ureg.m,
                    "xmax": 50 * ureg.m,
                    "ymin": -50 * ureg.m,
                    "ymax": 50 * ureg.m,
                    "z": 30 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1},
    )

    result = eradiate.run(exp)

    test = Chi2Test(
        name=f"{session_timestamp:%Y%m%d-%H%M%S}-het01.nc",
        value=result,
        reference="tests/regression_test_references/het01_brfpp_ref.nc",
        threshold=0.05,
        archive_dir=artefact_dir,
    )

    assert test.run()



import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg


@pytest.mark.slow
@pytest.mark.parametrize("bottom", [0.0, 1.0, 1.0])
@pytest.mark.parametrize("tau_ref", [0.1, 1.0, 10.0])
def test_heterogeneous_atmosphere_expansion_particle_layer(
    mode_ckd_double, bottom, tau_ref, ert_seed_state
):
    """
    Perfect single-component HeterogeneousAtmosphere expansion (particle layer)
    ===========================================================================

    This test case checks if a single-component HeterogeneousAtmosphere
    container expands perfectly as its component. This is assessed by checking
    if the results provided by two scenes presumably identical are indeed the
    same.

    Rationale
    ---------

    Run an AtmosphereExperiment with two different atmosphere definitions:

    1. A single particle layer.
    2. Heterogeneous atmosphere that contains a single particle layer.

    Expected behaviour
    ------------------

    The two experiments produce *exactly* the same results.
    """
    spp = 1000

    # Measure
    measure = {
        "type": "distant",
        "id": "measure",
        "construct": "hplane",
        "zeniths": np.arange(-90, 91, 15),
        "azimuth": 0.0,
        "srf": eradiate.scenes.spectra.MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm),
    }

    # Particle layer only
    bottom = bottom * ureg.km
    top = bottom + 1.0 * ureg.km
    geometry = {
        "type": "plane_parallel",
        "ground_altitude": bottom,
        "toa_altitude": top,
    }
    layer = {
        "type": "particle_layer",
        "bottom": bottom,
        "top": top,
        "tau_ref": tau_ref,
    }
    exp1 = eradiate.experiments.AtmosphereExperiment(
        geometry=geometry,
        atmosphere=layer,
        measures=[measure],
    )
    ert_seed_state.reset()
    eradiate.run(exp1, seed_state=ert_seed_state, spp=spp)
    results1 = exp1.results["measure"]["radiance"].values

    # Heterogeneous atmosphere with a particle layer
    exp2 = eradiate.experiments.AtmosphereExperiment(
        geometry=geometry,
        atmosphere={
            "type": "heterogeneous",
            "particle_layers": [layer],
        },
        measures=[measure],
    )
    ert_seed_state.reset()
    eradiate.run(exp2, seed_state=ert_seed_state, spp=spp)
    results2 = exp2.results["measure"]["radiance"].values

    np.testing.assert_equal(results1, results2)


def test_heterogeneous_atmosphere_expansion_molecular_atmosphere(
    mode_ckd_double, ert_seed_state, atmosphere_us_standard_ckd
):
    """
    Perfect single-component HeterogeneousAtmosphere expansion (molecular atmosphere)
    =================================================================================

    This test case checks if a single-component HeterogeneousAtmosphere
    container expands perfectly as its component. This is assessed by checking
    if the results provided by two scenes presumably identical are indeed the
    same.

    Rationale
    ---------

    Run an AtmosphereExperiment with two different atmosphere definitions:

    1. A molecular atmosphere.
    2. Heterogeneous atmosphere that contains a molecular atmosphere only.

    Expected behaviour
    ------------------

    The two experiments produce *exactly* the same results.
    """
    spp = 1000

    # measure
    measure = {
        "type": "distant",
        "id": "measure",
        "construct": "hplane",
        "zeniths": np.arange(-90, 90, 15),
        "azimuth": 0.0,
        "srf": {
            "type": "multi_delta",
            "wavelengths": 550.0 * ureg.nm,
        },
    }

    # Non-absorbing molecular atmosphere
    molecular_atmosphere = atmosphere_us_standard_ckd
    molecular_atmosphere.update({"has_absorption": False})
    exp1 = eradiate.experiments.AtmosphereExperiment(
        atmosphere=molecular_atmosphere,
        measures=[measure],
    )
    ert_seed_state.reset()
    eradiate.run(exp1, seed_state=ert_seed_state, spp=spp)
    results1 = exp1.results["measure"]["radiance"].values

    # Heterogeneous atmopshere with a non-absorbing molecular atmosphere
    exp2 = eradiate.experiments.AtmosphereExperiment(
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": molecular_atmosphere,
        },
        measures=[measure],
    )
    ert_seed_state.reset()
    eradiate.run(exp2, seed_state=ert_seed_state)
    results2 = exp2.results["measure"]["radiance"].values

    np.testing.assert_equal(results1, results2)



import mitsuba as mi
import numpy as np

import eradiate
from eradiate import KernelContext
from eradiate import unit_context_kernel as uck
from eradiate import unit_registry as ureg
from eradiate.scenes.spectra import AirScatteringCoefficientSpectrum
from eradiate.test_tools.types import check_scene_element


def test_air_scattering_coefficient_construct(modes_all):
    # Construction without argument succeeds
    s = AirScatteringCoefficientSpectrum()
    assert s


def test_air_scattering_coefficient_eval(modes_all_double):
    # The spectrum evaluates correctly (reference values computed manually)
    expected = 0.0114934 / ureg.km  # value at 550 nm (default wavelength)

    s = AirScatteringCoefficientSpectrum()
    ctx = KernelContext()

    value = s.eval(ctx.si)
    assert np.allclose(value, expected)


def test_air_scattering_coefficient_kernel_dict(modes_all):
    s = AirScatteringCoefficientSpectrum()

    with uck.override(length="m"):
        mi_wrapper = check_scene_element(s, mi.Texture)

    if eradiate.mode().is_mono:
        expected = ureg.convert(0.0114934, "km^-1", "m^-1")

    elif eradiate.mode().is_ckd:
        expected = ureg.convert(0.0114968, "km^-1", "m^-1")

    else:
        raise ValueError(f"no reference value for mode {eradiate.mode()}")

    assert np.isclose(mi_wrapper.parameters["value"], expected)



from __future__ import annotations

from abc import ABC

import attrs

from ..core import NodeSceneElement
from ..._factory import Factory
from ...attrs import documented, get_doc, parse_docs

illumination_factory = Factory()
illumination_factory.register_lazy_batch(
    [
        ("_astro_object.AstroObjectIllumination", "astro_object", {}),
        ("_constant.ConstantIllumination", "constant", {}),
        ("_directional.DirectionalIllumination", "directional", {}),
        ("_spot.SpotIllumination", "spot", {}),
    ],
    cls_prefix="eradiate.scenes.illumination",
)


@parse_docs
@attrs.define(eq=False, slots=False)
class Illumination(NodeSceneElement, ABC):
    """
    Abstract base class for all illumination scene elements.

    Notes
    -----
    * This class is to be used as a mixin.
    """

    id: str | None = documented(
        attrs.field(
            default="illumination",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(NodeSceneElement, "id", "doc"),
        type=get_doc(NodeSceneElement, "id", "type"),
        init_type=get_doc(NodeSceneElement, "id", "init_type"),
        default='"illumination"',
    )



import mitsuba as mi
import pytest

from eradiate import KernelContext
from eradiate.exceptions import TraversalError
from eradiate.scenes.core import Scene, traverse
from eradiate.scenes.shapes import RectangleShape, Shape
from eradiate.scenes.surface import BasicSurface
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs, expected_shape, expected_traversal_param_keys",
    [
        ({}, None, TraversalError),
        (
            {"shape": {"type": "rectangle"}},
            RectangleShape,
            {"surface_bsdf.reflectance.value"},
        ),
    ],
    ids=[
        "noargs",
        "shape",
    ],
)
def test_basic_surface_construct(
    modes_all_double, kwargs, expected_shape, expected_traversal_param_keys
):
    surface = BasicSurface(**kwargs)

    if expected_shape is None:
        assert surface.shape is None
    elif issubclass(expected_shape, Shape):
        assert isinstance(surface.shape, expected_shape)
    else:
        raise NotImplementedError

    if isinstance(expected_traversal_param_keys, set):
        template, params = traverse(surface)

        # Scene element is composite: template has not "type" key
        assert "type" not in template
        # Parameter map keys are fetched recursively
        assert set(params.keys()) == expected_traversal_param_keys

        # When enclosed in a Scene, the surface can be traversed
        scene = Scene(objects={"surface": surface})
        template, params = traverse(scene)
        kernel_dict = template.render(KernelContext())
        assert isinstance(mi.load_dict(kernel_dict), mi.Scene)

    elif isinstance(expected_traversal_param_keys, type) and issubclass(
        expected_traversal_param_keys, TraversalError
    ):
        with pytest.raises(
            expected_traversal_param_keys,
            match="A 'BasicSurface' cannot be traversed if its 'shape' field is unset",
        ):
            traverse(surface)
    else:
        raise NotImplementedError


def test_basic_surface_kernel_dict(mode_mono):
    surface = BasicSurface(shape={"type": "rectangle"})
    check_scene_element(surface)



import os
import tempfile

import attr
import numpy as np
import pytest

from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.scenes.biosphere._leaf_cloud import (
    LeafCloud,
    _leaf_cloud_orientations,
    _leaf_cloud_positions_cuboid,
    _leaf_cloud_positions_cuboid_avoid_overlap,
    _leaf_cloud_positions_cylinder,
    _leaf_cloud_positions_ellipsoid,
    _leaf_cloud_radii,
    _sample_lad,
)
from eradiate.scenes.core import traverse
from eradiate.test_tools.types import check_scene_element


@pytest.fixture(scope="function")
def rng():
    yield np.random.default_rng(seed=12345)


@pytest.fixture(scope="function")
def rng_mock():
    @attr.s
    class RngMock:
        values = attr.ib(converter=np.atleast_1d)

        def random(self, n=1):
            selector = np.array([i % len(self.values) for i in range(n)])
            return self.values[selector].squeeze()

    yield RngMock


@pytest.fixture(scope="module")
def tempfile_leaves():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_leaves.txt")
        with open(filename, "w") as tf:
            tf.write("0.100 8.864 9.040 1.878 -0.314 0.025 0.949\n")
            tf.write("0.100 9.539 -10.463 0.627 0.489 -0.276 0.828\n")
            tf.write("0.100 -2.274 -9.204 0.797 0.618 0.184 0.764\n")
            tf.write("0.100 -9.957 -4.971 0.719 -0.066 0.100 0.993\n")
            tf.write("0.100 5.339 9.153 0.500 0.073 -0.294 0.953\n")
        yield filename


# -- Basic function tests ------------------------------------------------------


def test_sample_lad(rng_mock):
    """
    Unit tests for :func:`_sample_lad`.
    The RNG Mock lets us test assert the probability of sampling a certain
    value from the Goel and Strebel LAD function.

    The a member of the mock represents the theta sample and is mapped
    to the [0, pi/2.] range.
    The b member represents the rejection sample value.
    """

    rng = rng_mock([0.1, 0.45])
    assert np.allclose(_sample_lad(2.531, 1.096, rng), np.pi / 20.0)

    rng = rng_mock([0.9, 0.45])
    assert np.allclose(_sample_lad(1.096, 2.531, rng), np.pi / 2.0 * 0.9)


def test_leaf_cloud_positions_cuboid(rng):
    """Unit tests for :func:`_leaf_cloud_positions_cuboid`."""
    positions = _leaf_cloud_positions_cuboid(1, 1.0 * ureg.m, 1.0 * ureg.m, rng)
    assert positions.shape == (1, 3)
    assert np.allclose(positions, [-0.27266398, -0.18324166, 0.79736546] * ureg.m)


def test_leaf_cloud_positions_cuboid_avoid_overlap(rng):
    """Unit tests for :func:`_leaf_cloud_positions_cuboid_avoid_overlap`."""
    positions = _leaf_cloud_positions_cuboid_avoid_overlap(
        1000, 1.0 * ureg.m, 1.0 * ureg.m, 1.0 * ureg.cm, 1e3, rng
    )
    assert positions.shape == (1000, 3)


def test_leaf_cloud_positions_ellipsoid(rng):
    """Unit tests for :func:`_leaf_cloud_positions_ellipsoid`."""
    positions = _leaf_cloud_positions_ellipsoid(
        1, rng, 1.0 * ureg.m, 1.0 * ureg.m, 1.0 * ureg.m
    )
    assert positions.shape == (1, 3)
    assert np.allclose(positions, [-0.54532796, -0.36648332, 0.59473091] * ureg.m)


def test_leaf_cloud_positions_cylinder(rng):
    """Unit tests for :func:`_leaf_cloud_positions_cylinder`."""
    positions = _leaf_cloud_positions_cylinder(1, 1.0 * ureg.m, 1.0 * ureg.m, rng)
    assert positions.shape == (1, 3)
    assert np.allclose(positions, [0.04495472, 0.3135521, 0.79736546] * ureg.m)


def test_leaf_cloud_orientations(rng_mock):
    """Unit tests for :func:`_leaf_cloud_orientations`."""
    rng = rng_mock([0.1, 0.45])

    orientations = _leaf_cloud_orientations(1, 2.531, 1.096, rng)
    assert orientations.shape == (1, 3)
    assert np.allclose(orientations, [0.1265581, 0.09194987, 0.98768834])


def test_leaf_cloud_radii():
    radii = _leaf_cloud_radii(1, 10.0 * ureg.cm)
    assert radii.shape == (1,)
    assert np.allclose(radii, 10.0 * ureg.cm)


# ------------------------------------------------------------------------------
#                                 LeafCloud tests
# ------------------------------------------------------------------------------


def test_leaf_cloud_instantiate(mode_mono):
    """Unit tests for :class:`LeafCloud`'s default constructor."""

    # Empty constructor does not raise
    cloud = LeafCloud()
    check_scene_element(cloud)

    # Now with more sensible values, also check template and parameter map
    cloud = LeafCloud(
        id="cloud",
        leaf_positions=[[0, 0, 0]],
        leaf_orientations=[[0, 0, 1]],
        leaf_radii=[0.1],
    )
    check_scene_element(cloud)
    template, params = traverse(cloud)
    assert set(template.keys()) == {
        "bsdf_cloud.type",
        "bsdf_cloud.reflectance.type",
        "bsdf_cloud.reflectance.value",
        "bsdf_cloud.transmittance.type",
        "bsdf_cloud.transmittance.value",
        "cloud_leaf_0.type",
        "cloud_leaf_0.bsdf.type",
        "cloud_leaf_0.bsdf.id",
        "cloud_leaf_0.to_world",
    }
    assert set(params.keys()) == {
        "bsdf_cloud.reflectance.value",
        "bsdf_cloud.transmittance.value",
    }

    # Incorrect dimensionality raises
    with pytest.raises(ValueError):
        LeafCloud(
            leaf_positions=[0, 0, 0],
            leaf_orientations=[0, 0, 1],
            leaf_radii=[0.1] * 3,
        )


@pytest.mark.parametrize(
    "leaf_positions, xyz, expected_leaf_positions",
    [
        (
            [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]],
            [0, 0, 1],
            [[0, 0, 1], [0, 0, 2], [0, 1, 1], [1, 0, 1]],
        ),
        (
            [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]],
            [[0, 0, 1]] + [[0, 0, 0]] * 3,
            [[0, 0, 1], [0, 0, 1], [0, 1, 0], [1, 0, 0]],
        ),
    ],
)
def test_leaf_cloud_translated(mode_mono, leaf_positions, xyz, expected_leaf_positions):
    """Unit tests for :meth:`LeafCloud.translated`."""

    leaf_positions_size = np.array(leaf_positions).size
    leaf_orientations = [[0, 0, 1]] * (leaf_positions_size // 3)
    leaf_radii = [0.1] * (leaf_positions_size // 3)

    cloud = LeafCloud(
        leaf_positions=leaf_positions,
        leaf_orientations=leaf_orientations,
        leaf_radii=leaf_radii,
    )

    # Translated object is a copy
    translated_leaf_cloud = cloud.translated(xyz * ureg.m)
    assert translated_leaf_cloud is not cloud

    # Translated object has offset leaf positions
    assert np.allclose(
        translated_leaf_cloud.leaf_positions,
        np.array(expected_leaf_positions) * ureg.m,
    )

    # Shapes do not match: translated() raises
    with pytest.raises(ValueError):
        cloud.translated([[1, 0, 0]] * 2 * ureg.m)


@pytest.mark.slow
def test_leaf_cloud_generate(mode_mono):
    """Test the instantiation of the leaf cloud class from parameters"""

    # Try constructing leaf clouds with various param combinations
    # n_leaves, leaf_radius, l_horizontal, l_vertical
    cloud = LeafCloud.cuboid(
        n_leaves=1000,
        leaf_radius=10.0 * ureg.cm,
        l_horizontal=30.0 * ureg.m,
        l_vertical=3.0 * ureg.m,
    )
    check_scene_element(cloud)
    assert cloud.n_leaves() == 1000
    # Cloud covers the expected extent
    assert np.allclose(
        np.min(cloud.leaf_positions, axis=0).m_as("m"), [-15, -15, 0], atol=0.05
    )
    assert np.allclose(
        np.max(cloud.leaf_positions, axis=0).m_as("m"), [15, 15, 3], atol=0.05
    )

    # lai, leaf_radius, l_horizontal, l_vertical
    cloud = LeafCloud.cuboid(
        lai=3.0,
        leaf_radius=15.0 * ureg.cm,
        l_horizontal=10.0 * ureg.m,
        l_vertical=3.0 * ureg.m,
    )
    check_scene_element(cloud)
    assert cloud.n_leaves() == 4244
    assert np.allclose(
        np.min(cloud.leaf_positions, axis=0).m_as("m"), [-5, -5, 0], atol=0.05
    )
    assert np.allclose(
        np.max(cloud.leaf_positions, axis=0).m_as("m"), [5, 5, 3], atol=0.05
    )

    # lai, leaf_radius, l_horizontal, hdo, hvr
    cloud = LeafCloud.cuboid(
        lai=2.0,
        leaf_radius=10.0 * ureg.cm,
        l_horizontal=10.0 * ureg.m,
        hdo=10 * ureg.cm,
        hvr=0.05,
    )
    check_scene_element(cloud)
    assert cloud.n_leaves() == 6366
    assert np.allclose(
        np.min(cloud.leaf_positions, axis=0).m_as("m"), [-5, -5, 0], atol=0.05
    )
    assert np.allclose(
        np.max(cloud.leaf_positions, axis=0).m_as("m"), [5, 5, 1.272], atol=0.05
    )

    # n_leaves, leaf_radius, l_horizontal, l_vertical, collision check
    cloud = LeafCloud.cuboid(
        n_leaves=1000,
        leaf_radius=10.0 * ureg.cm,
        l_horizontal=30.0 * ureg.m,
        l_vertical=3.0 * ureg.m,
        avoid_overlap=True,
    )
    check_scene_element(cloud)
    assert cloud.n_leaves() == 1000
    assert np.allclose(
        np.min(cloud.leaf_positions.m_as("m"), axis=0), [-15, -15, 0], atol=0.05
    )
    assert np.allclose(
        np.max(cloud.leaf_positions.m_as("m"), axis=0), [15, 15, 3], atol=0.05
    )


def test_leaf_cloud_from_file(mode_mono, tempfile_leaves):
    """Unit testing for :meth:`LeafCloud.from_file`."""

    # A LeafCloud instance can be loaded from a file on the hard drive
    leaf_cloud = LeafCloud.from_file(tempfile_leaves)
    assert len(leaf_cloud.leaf_positions) == 5
    assert np.allclose(leaf_cloud.leaf_radii, 0.1 * ureg.m)

    # Produced kernel dict is valid
    check_scene_element(leaf_cloud)


def test_leaf_cloud_kernel_dict(mode_mono):
    """Partial unit testing for :meth:`LeafCloud.kernel_dict`."""

    leaf_cloud_id = "leaf_cloud"
    leaf_cloud = LeafCloud(
        id=leaf_cloud_id,
        leaf_positions=[[0, 0, 0], [1, 1, 1]],
        leaf_orientations=[[1, 0, 0], [0, 1, 0]],
        leaf_radii=[0.1, 0.1],
        leaf_reflectance=0.5,
        leaf_transmittance=0.5,
    )

    template, params = traverse(leaf_cloud)

    # The BSDF is bilambertian with the parameters we initially set
    kernel_dict = template.render(ctx=KernelContext())
    assert kernel_dict[f"bsdf_{leaf_cloud_id}"] == {
        "type": "bilambertian",
        "reflectance": {"type": "uniform", "value": 0.5},
        "transmittance": {"type": "uniform", "value": 0.5},
    }

    # Leaves are disks
    for shape_key in [f"{leaf_cloud_id}_leaf_0", f"{leaf_cloud_id}_leaf_1"]:
        assert kernel_dict[shape_key]["type"] == "disk"

    # Kernel dict is valid
    check_scene_element(leaf_cloud)


def test_surface_area(mode_mono):
    """Unit testing for :meth:`LeafCloud.surface_area`."""
    cloud = LeafCloud(
        leaf_positions=[[0, 0, 0], [1, 1, 1]],
        leaf_orientations=[[1, 0, 0], [0, 1, 0]],
        leaf_radii=[0.1, 0.1],
    )

    assert cloud.surface_area() == (2.0 * np.pi * 0.1**2) * (ureg.m**2)



import pytest
from pinttr.exceptions import UnitsError

from eradiate import unit_registry as ureg
from eradiate.scenes.spectra import UniformSpectrum, spectrum_factory


@pytest.mark.parametrize(
    "tested, expected",
    [
        (
            {"type": "uniform", "value": 1.0},
            UniformSpectrum(quantity="radiance", value=1.0),
        ),
        (
            1.0,
            UniformSpectrum(quantity="radiance", value=1.0),
        ),
        (
            ureg.Quantity(1e6, "W/km^2/sr/nm"),
            UniformSpectrum(quantity="radiance", value=1.0),
        ),
        (
            1,
            UniformSpectrum(quantity="radiance", value=1.0),
        ),
        (
            ureg.Quantity(1, "W/m^2/nm"),
            UnitsError,
        ),
    ],
    ids=["dict", "float", "quantity", "int", "wrong_units"],
)
def test_converter(mode_mono, tested, expected):
    """
    Tests for the Spectrum factory's conversion protocol.
    """

    if isinstance(expected, UniformSpectrum):
        s = spectrum_factory.converter("radiance")(tested)
        assert (s.quantity == expected.quantity) and (
            s.value == expected.value
        ), f"Failed: converted value is {s}"

    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            spectrum_factory.converter("radiance")(tested)

    else:
        raise RuntimeError



import drjit as dr
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.atmosphere import HomogeneousAtmosphere
from eradiate.scenes.core import traverse
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"phase": {"type": "isotropic"}},
        {"phase": {"type": "hg"}},
        {"geometry": "spherical_shell"},
    ],
    ids=[
        "noargs",
        "phase_isotropic",
        "phase_hg",
        "geometry_spherical_shell",
    ],
)
def test_homogeneous_atmosphere_construct(modes_all_double, kwargs):
    atmosphere = HomogeneousAtmosphere(**kwargs)
    check_scene_element(atmosphere)


def test_homogeneous_atmosphere_params(mode_mono):
    atmosphere = HomogeneousAtmosphere(
        phase={
            "type": "hg",
            "g": {
                "type": "interpolated",
                "wavelengths": [400, 700],
                "values": [0.0, 0.8],
            },
        }
    )

    _, umap_template = traverse(atmosphere)
    mi_wrapper = check_scene_element(atmosphere)

    # Phase function parameters are exposed at highest level
    assert "medium_atmosphere.phase_function.g" in umap_template
    assert "medium_atmosphere.phase_function.g" in mi_wrapper.parameters.keys()
    # Volume data source parameters are exposed at highest level
    assert "medium_atmosphere.sigma_t.value.value" in umap_template
    assert "medium_atmosphere.albedo.value.value" in umap_template


@pytest.mark.parametrize(
    "kwargs, expected",
    [
        (
            {
                "geometry": {
                    "type": "plane_parallel",
                    "width": 1.0 * ureg.km,
                    "ground_altitude": 0.0 * ureg.km,
                    "toa_altitude": 1.0 * ureg.km,
                }
            },
            {
                "bbox_min": [-500.0, -500.0, -10.0],
                "bbox_max": [500.0, 500.0, 1000.0],
                "atmosphere_shape_type_name": "Cube",
            },
        ),
        (
            {
                "geometry": {
                    "type": "spherical_shell",
                    "planet_radius": 1.0 * ureg.km,
                    "ground_altitude": 0.0 * ureg.km,
                    "toa_altitude": 1.0 * ureg.km,
                },
            },
            {
                "bbox_min": [-2000.0, -2000.0, -2000.0],
                "bbox_max": [2000.0, 2000.0, 2000.0],
                "atmosphere_shape_type_name": "Sphere",
            },
        ),
    ],
    ids=[
        "plane_parallel",
        "spherical_shell",
    ],
)
def test_homogeneous_atmosphere_geometry(mode_mono, kwargs, expected):
    atmosphere = HomogeneousAtmosphere(**kwargs)
    mi_wrapper = check_scene_element(atmosphere)

    # Check scene shape type
    mi_shape_type_name = mi_wrapper.obj.shapes()[0].class_().name()
    assert mi_shape_type_name == expected["atmosphere_shape_type_name"]

    # Check scene bounding box
    bbmin = mi_wrapper.obj.bbox().min
    bbmax = mi_wrapper.obj.bbox().max
    assert dr.allclose(bbmin, expected["bbox_min"])
    assert dr.allclose(bbmax, expected["bbox_max"])
